"""Test correctness of Drugs@FDA ETL methods."""

import isodate
import pytest

from therapy.etl.drugsatfda import DrugsAtFDA
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def drugsatfda(test_source):
    """Provide test Drugs@FDA query endpoint"""
    return test_source(DrugsAtFDA)


@pytest.fixture(scope="module")
def everolimus() -> Therapy:
    """Build afinitor test fixture."""
    return Therapy(
        label="EVEROLIMUS",
        concept_id="drugsatfda.nda:022334",
        xrefs=[
            "rxcui:1119400",
            "rxcui:1119402",
            "rxcui:1308428",
            "rxcui:1308430",
            "rxcui:1308432",
            "rxcui:1310138",
            "rxcui:1310144",
            "rxcui:1310147",
            "rxcui:845507",
            "rxcui:845512",
            "rxcui:845515",
            "rxcui:845518",
            "rxcui:998189",
            "rxcui:998191",
        ],
        associated_with=[
            "ndc:0078-0566",
            "ndc:0078-0566",
            "ndc:0078-0567",
            "ndc:0078-0594",
            "ndc:0078-0620",
            "ndc:0078-0626",
            "ndc:0078-0627",
            "ndc:0078-0628",
            "spl:a2136e71-186e-4bfc-901b-5ed9ca048a1f",
            "unii:9HW64Q8G6G",
        ],
        approval_ratings=["fda_prescription"],
        trade_names=["AFINITOR"],
    )


@pytest.fixture(scope="module")
def dactinomycin() -> Therapy:
    """Build cosmegen test fixture."""
    return Therapy(
        label="DACTINOMYCIN",
        concept_id="drugsatfda.nda:050682",
        xrefs=["rxcui:105569", "rxcui:239179"],
        associated_with=[
            "ndc:55292-811",
            "ndc:66993-489",
            "spl:c76d44e7-3869-4e8b-93cf-b031f6f62c0a",
            "spl:2afdf978-8e90-40e9-872f-9f8775dcd41c",
            "unii:1CC1JFE158",
        ],
        approval_ratings=["fda_prescription"],
        trade_names=["COSMEGEN"],
    )


@pytest.fixture(scope="module")
def cisplatin() -> list[Therapy]:
    """Build cisplatin test fixtures"""
    return [
        Therapy(
            label="CISPLATIN",
            concept_id="drugsatfda.anda:074656",
            xrefs=["rxcui:309311"],
            associated_with=[
                "ndc:0703-5747",
                "ndc:0703-5748",
                "spl:adf5773e-9095-4cb4-a90f-72cbf82f4493",
                "unii:Q20Q21Q62J",
            ],
            approval_ratings=["fda_prescription"],
        ),
        Therapy(
            label="CISPLATIN",
            concept_id="drugsatfda.anda:075036",
            xrefs=["rxcui:309311"],
            associated_with=[
                "ndc:0143-9504",
                "ndc:0143-9505",
                "spl:2c569ef0-588f-4828-8b2d-03a2120c9b4c",
                "unii:Q20Q21Q62J",
            ],
            approval_ratings=["fda_prescription"],
        ),
        Therapy(
            label="CISPLATIN",
            concept_id="drugsatfda.anda:074735",
            xrefs=["rxcui:309311"],
            associated_with=[
                "ndc:63323-103",
                "spl:9b008181-ab66-db2f-e053-2995a90aad57",
                "unii:Q20Q21Q62J",
            ],
            approval_ratings=["fda_prescription"],
        ),
        Therapy(
            label="CISPLATIN",
            concept_id="drugsatfda.anda:206774",
            xrefs=["rxcui:309311"],
            associated_with=[
                "ndc:16729-288",
                "ndc:68001-283",
                "spl:c3ddc4a5-9f1b-a8ee-e053-2a95a90a2265",
                "spl:c43de769-d6d8-3bb9-e053-2995a90a5aa2",
                "unii:Q20Q21Q62J",
            ],
            approval_ratings=["fda_prescription"],
        ),
        Therapy(
            label="CISPLATIN",
            concept_id="drugsatfda.anda:207323",
            xrefs=["rxcui:309311"],
            associated_with=[
                "ndc:25021-253",
                "ndc:72266-252",
                "ndc:72266-253",
                "ndc:68083-162",
                "ndc:68083-163",
                "ndc:70860-206",
                "spl:01c7a680-ee0d-42da-85e8-8d56c6fe7006",
                "spl:0219ee77-eb38-b81f-e063-6394a90a5ca4",
                "spl:64bcce1a-6e31-4e73-8da5-11aa9e890da2",
                "spl:89007399-3c34-40d2-8068-a0b8e09cbef8",
                "unii:Q20Q21Q62J",
            ],
            approval_ratings=["fda_prescription"],
        ),
        Therapy(
            label="CISPLATIN",
            concept_id="drugsatfda.nda:018057",
            xrefs=["rxcui:1736854", "rxcui:309311"],
            associated_with=[
                "ndc:44567-509",
                "ndc:44567-510",
                "ndc:44567-511",
                "ndc:44567-530",
                "spl:a66eda32-1164-439a-ac8e-73138365ec06",
                "spl:dd45d777-d4c1-40ee-b4f0-c9e001a15a8c",
                "unii:Q20Q21Q62J",
            ],
            trade_names=["PLATINOL-AQ", "PLATINOL"],
        ),
    ]


@pytest.fixture(scope="module")
def fenortho() -> list[Therapy]:
    """Provide fenortho fixture. Tests for correct handling of conflicting
    approval_ratings values.
    """
    return [
        Therapy(
            label="FENOPROFEN CALCIUM",
            concept_id="drugsatfda.anda:072267",
            xrefs=["rxcui:310291", "rxcui:351398"],
            associated_with=[
                "ndc:42195-471",
                "ndc:42195-688",
                "unii:0X2CW1QABJ",
                "spl:079e140a-b3a2-1c2e-e063-6394a90a8491",
                "spl:079e260a-4bee-a423-e063-6394a90aaf0d",
            ],
            approval_ratings=["fda_prescription"],
            trade_names=["NALFON"],
        ),
        Therapy(
            label="FENOPROFEN CALCIUM",
            concept_id="drugsatfda.nda:017604",
            xrefs=[
                "rxcui:197694",
                "rxcui:197695",
                "rxcui:260323",
                "rxcui:858116",
                "rxcui:858118",
            ],
            associated_with=[
                "ndc:0276-0501",
                "ndc:0276-0502",
                "ndc:0276-0503",
                "ndc:15014-400",
                "ndc:42195-100",
                "ndc:42195-308",
                "ndc:42195-600",
                "spl:d8a6c130-bc1d-01a4-e053-2995a90afd52",
                "spl:f0305a11-4494-e237-e053-2995a90ad12e",
                "unii:0X2CW1QABJ",
                "spl:079cb594-ac07-14e3-e063-6394a90aeb77",
                "spl:079db405-aab1-a400-e063-6294a90a1489",
            ],
            trade_names=["NALFON"],
        ),
        Therapy(
            concept_id="drugsatfda.anda:214475",
            label="FENOPROFEN CALCIUM",
            xrefs=["rxcui:858116"],
            associated_with=[
                "spl:b3a58d34-7780-4056-b144-bad674c51494",
                "unii:0X2CW1QABJ",
                "ndc:16571-688",
            ],
            approval_ratings=["fda_prescription"],
        ),
    ]


def test_everolimus(drugsatfda, compare_records, everolimus):
    """Test everolimus/afinitor"""
    concept_id = "drugsatfda.nda:022334"

    # test concept ID
    response = drugsatfda.search(concept_id)
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], everolimus)

    response = drugsatfda.search("drugsatfda.anda:022334")
    assert response.match_type == MatchType.NO_MATCH

    # test label
    response = drugsatfda.search("everolimus")
    assert response.match_type == MatchType.LABEL
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], everolimus)

    # test trade name
    response = drugsatfda.search("afinitor")
    assert response.match_type == MatchType.TRADE_NAME
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], everolimus)

    # test xref
    response = drugsatfda.search("rxcui:998191")
    assert response.match_type == MatchType.XREF
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], everolimus)

    response = drugsatfda.search("rxcui:845507")
    assert response.match_type == MatchType.XREF
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], everolimus)

    response = drugsatfda.search("rxcui:998189")
    assert response.match_type == MatchType.XREF
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], everolimus)

    response = drugsatfda.search("rxcui:1308432")
    assert response.match_type == MatchType.XREF
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], everolimus)

    # test assoc_with
    response = drugsatfda.search("unii:9HW64Q8G6G")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], everolimus)


def test_dactinomycin(drugsatfda, compare_records, dactinomycin):
    """Test dactinomycin"""
    concept_id = "drugsatfda.nda:050682"

    # test concept ID
    response = drugsatfda.search(concept_id)
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], dactinomycin)

    # test label
    response = drugsatfda.search("DACTINOMYCIN")
    assert response.match_type == MatchType.LABEL
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], dactinomycin)

    # test trade name
    response = drugsatfda.search("cosmegen")
    assert response.match_type == MatchType.TRADE_NAME
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], dactinomycin)

    # test xref
    response = drugsatfda.search("rxcui:105569")
    assert response.match_type == MatchType.XREF
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], dactinomycin)

    response = drugsatfda.search("rxcui:239179")
    assert response.match_type == MatchType.XREF
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], dactinomycin)

    # test assoc_with
    response = drugsatfda.search("unii:1cc1jfe158")
    assert response.match_type == MatchType.XREF
    records = [r for r in response.records if r.concept_id == concept_id]
    compare_records(records[0], dactinomycin)


def test_cisplatin(drugsatfda, compare_records, cisplatin):
    """Test cisplatin"""
    # test concept IDs
    concept_id = "drugsatfda.anda:074656"
    response = drugsatfda.search(concept_id)
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(
        response.records[0], next(c for c in cisplatin if c.concept_id == concept_id)
    )

    concept_id = "drugsatfda.anda:075036"
    response = drugsatfda.search(concept_id)
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(
        response.records[0],
        next(c for c in cisplatin if c.concept_id.lower() == concept_id.lower()),
    )

    concept_id = "drugsatfda.anda:074735"
    response = drugsatfda.search(concept_id)
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(
        response.records[0], next(c for c in cisplatin if c.concept_id == concept_id)
    )

    concept_id = "drugsatfda.anda:206774"
    response = drugsatfda.search(concept_id)
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(
        response.records[0], next(c for c in cisplatin if c.concept_id == concept_id)
    )

    concept_id = "drugsatfda.anda:207323"
    response = drugsatfda.search(concept_id)
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(
        response.records[0], next(c for c in cisplatin if c.concept_id == concept_id)
    )

    # test label
    response = drugsatfda.search("cisplatin")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 6
    for r in response.records:
        fixture = next(c for c in cisplatin if r.concept_id == c.concept_id)
        compare_records(r, fixture)

    # test xref
    response = drugsatfda.search("rxcui:309311")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 6
    for r in response.records:
        fixture = next(c for c in cisplatin if r.concept_id == c.concept_id)
        compare_records(r, fixture)

    # test assoc_with
    response = drugsatfda.search("unii:q20q21q62j")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 6
    for r in response.records:
        fixture = next(c for c in cisplatin if r.concept_id == c.concept_id)
        compare_records(r, fixture)


def test_fenortho(fenortho, compare_records, drugsatfda):
    """Test fenortho."""
    response = drugsatfda.search("drugsatfda.anda:072267")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], fenortho[0])

    response = drugsatfda.search("drugsatfda.nda:017604")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], fenortho[1])

    response = drugsatfda.search("fenoprofen calcium")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 3
    for r in response.records:
        fixture = next(f for f in fenortho if r.concept_id == f.concept_id)
        compare_records(r, fixture)

    response = drugsatfda.search("rxcui:310291")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], fenortho[0])

    response = drugsatfda.search("unii:0X2CW1QABJ")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 3
    for r in response.records:
        fixture = next(f for f in fenortho if r.concept_id == f.concept_id)
        compare_records(r, fixture)


def test_other_parameters(drugsatfda):
    """Test other design decisions that aren"t captured in the defined
    fixtures.
    """
    # test dropping BLA records
    response = drugsatfda.search("drugsatfda:BLA020725")
    assert response.match_type == MatchType.NO_MATCH


def test_meta(drugsatfda):
    """Test correctness of source metadata."""
    response = drugsatfda.search("incoherent-string-of-text")
    assert response.source_meta_.data_license == "CC0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/publicdomain/zero/1.0/legalcode"
    )
    assert isodate.parse_date(response.source_meta_.version)
    assert (
        response.source_meta_.data_url
        == "https://open.fda.gov/apis/drug/drugsfda/download/"
    )
    assert response.source_meta_.rdp_url is None
    assert response.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "share_alike": False,
        "attribution": False,
    }



"""Test that the therapy normalizer works as intended for the HemOnc.org
source.
"""

import isodate
import pytest

from therapy.etl.hemonc import HemOnc
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def hemonc(test_source):
    """Provide test HemOnc query endpoint"""
    return test_source(HemOnc)


@pytest.fixture(scope="module")
def cisplatin():
    """Construct cisplatin fixture."""
    return Therapy(
        concept_id="hemonc:105",
        label="Cisplatin",
        aliases=[
            "cis-diamminedichloroplatinum III",
            "DDP",
            "cisplatinum",
            "cis-platinum",
            "DACP",
            "NSC 119875",
            "NSC-119875",
            "NSC119875",
            "CDDP",
        ],
        trade_names=[],
        xrefs=["rxcui:2555"],
        associated_with=[],
        approval_ratings=["hemonc_approved"],
        approval_year=["1978"],
        has_indication=[
            {
                "disease_id": "hemonc:569",
                "disease_label": "Bladder cancer",
                "normalized_disease_id": "ncit:C9334",
                "supplemental_info": {"regulatory_body": "FDA"},
            },
            {
                "disease_id": "hemonc:671",
                "disease_label": "Testicular cancer",
                "normalized_disease_id": "ncit:C7251",
                "supplemental_info": {"regulatory_body": "FDA"},
            },
            {
                "disease_id": "hemonc:645",
                "disease_label": "Ovarian cancer",
                "normalized_disease_id": "ncit:C7431",
                "supplemental_info": {"regulatory_body": "FDA"},
            },
        ],
    )


@pytest.fixture(scope="module")
def bendamustine():
    """Construct bendamustine fixture."""
    return Therapy(
        concept_id="hemonc:65",
        label="Bendamustine",
        aliases=[
            "CEP-18083",
            "cytostasan hydrochloride",
            "SyB L-0501",
            "SDX-105",
            "CEP 18083",
            "CEP18083",
            "SDX 105",
            "SDX105",
            "bendamustine hydrochloride",
            "bendamustin hydrochloride",
        ],
        xrefs=["rxcui:134547"],
        associated_with=[],
        trade_names=[
            "Belrapzo",
            "Bendamax",
            "Bendawel",
            "Bendeka",
            "Bendit",
            "Innomustine",
            "Leuben",
            "Levact",
            "Maxtorin",
            "MyMust",
            "Purplz",
            "Ribomustin",
            "Treakisym",
            "Treanda",
            "Vivimusta",
            "Xyotin",
        ],
        approval_ratings=["hemonc_approved"],
        approval_year=["2008"],
        has_indication=[
            {
                "disease_id": "hemonc:581",
                "disease_label": "Chronic lymphocytic leukemia",
                "normalized_disease_id": "ncit:C3163",
                "supplemental_info": {"regulatory_body": "FDA"},
            },
            {
                "disease_id": "hemonc:46094",
                "disease_label": "Indolent lymphoma",
                "normalized_disease_id": "ncit:C8504",
                "supplemental_info": {"regulatory_body": "FDA"},
            },
        ],
    )


@pytest.fixture(scope="module")
def degarelix():
    """Create fixture for degarelix drug."""
    return Therapy(
        label="Degarelix",
        concept_id="hemonc:151",
        aliases=[
            "degarelix acetate",
            "FE200486",
            "ASP3550",
            "ASP 3550",
            "ASP-3550",
            "FE 200486",
            "FE-200486",
        ],
        xrefs=["rxcui:475230"],
        associated_with=[],
        trade_names=["Firmagon", "Gonax"],
        approval_ratings=["hemonc_approved"],
        approval_year=["2008"],
        has_indication=[
            {
                "disease_id": "hemonc:658",
                "disease_label": "Prostate cancer",
                "normalized_disease_id": "ncit:C7378",
                "supplemental_info": {"regulatory_body": "FDA"},
            }
        ],
    )


@pytest.fixture(scope="module")
def filgrastim():
    """Create fixture for filgrastim drug (tests handling of deprecated brand name)"""
    return Therapy(
        label="Filgrastim-aafi",
        concept_id="hemonc:66258",
        aliases=[],
        xrefs=["rxcui:68442"],
        trade_names=["Nivestym"],
        approval_ratings=["hemonc_approved"],
        approval_year=["2018"],
    )


@pytest.fixture(scope="module")
def combo_drug():
    """Create fixture for `aspirin and dipyridamole`, a combo drug. Ensure that xref
    to rxcui:226716 isn't added.
    """
    return Therapy(
        label="Aspirin and dipyridamole",
        concept_id="hemonc:48",
        aliases=[],
        xrefs=[],
        trade_names=["Aggrenox"],
        approval_ratings=["hemonc_approved"],
        approval_year=["1999"],
    )


def test_concept_id_match(
    hemonc, compare_response, cisplatin, bendamustine, degarelix, filgrastim
):
    """Test that concept ID queries resolve to correct record."""
    response = hemonc.search("hemonc:105")
    compare_response(response, MatchType.CONCEPT_ID, cisplatin)

    response = hemonc.search("hemonc:65")
    compare_response(response, MatchType.CONCEPT_ID, bendamustine)

    response = hemonc.search("hemonc:151")
    compare_response(response, MatchType.CONCEPT_ID, degarelix)

    response = hemonc.search("hemonc:66258")
    compare_response(response, MatchType.CONCEPT_ID, filgrastim)


def test_label_match(
    hemonc, compare_response, cisplatin, bendamustine, degarelix, filgrastim
):
    """Test that label queries resolve to correct record."""
    response = hemonc.search("cisplatin")
    compare_response(response, MatchType.LABEL, cisplatin)

    response = hemonc.search("Bendamustine")
    compare_response(response, MatchType.LABEL, bendamustine)

    response = hemonc.search("DEGARELIX")
    compare_response(response, MatchType.LABEL, degarelix)

    response = hemonc.search("Filgrastim-aafi")
    compare_response(response, MatchType.LABEL, filgrastim)


def test_alias_match(hemonc, compare_response, cisplatin, bendamustine, degarelix):
    """Test that alias queries resolve to correct record."""
    response = hemonc.search("ddp")
    compare_response(response, MatchType.ALIAS, cisplatin)

    response = hemonc.search("dacp")
    compare_response(response, MatchType.ALIAS, cisplatin)

    response = hemonc.search("nsc 119875")
    compare_response(response, MatchType.ALIAS, cisplatin)

    response = hemonc.search("cep-18083")
    compare_response(response, MatchType.ALIAS, bendamustine)

    response = hemonc.search("bendamustine hydrochloride")
    compare_response(response, MatchType.ALIAS, bendamustine)

    response = hemonc.search("asp3550")
    compare_response(response, MatchType.ALIAS, degarelix)


def test_trade_name(hemonc, compare_response, bendamustine, degarelix):
    """Test that trade name queries resolve to correct record."""
    response = hemonc.search("bendamax")
    compare_response(response, MatchType.TRADE_NAME, bendamustine)

    response = hemonc.search("purplz")
    compare_response(response, MatchType.TRADE_NAME, bendamustine)

    response = hemonc.search("firmagon")
    compare_response(response, MatchType.TRADE_NAME, degarelix)

    # no trade names for records with > 20
    response = hemonc.search("platinol")
    assert response.match_type == MatchType.NO_MATCH


def test_xref_match(hemonc, compare_response, cisplatin, bendamustine, degarelix):
    """Test that xref query resolves to correct record."""
    response = hemonc.search("rxcui:2555")
    compare_response(response, MatchType.XREF, cisplatin)

    response = hemonc.search("rxcui:134547")
    compare_response(response, MatchType.XREF, bendamustine)

    response = hemonc.search("rxcui:475230")
    compare_response(response, MatchType.XREF, degarelix)


def test_combo_drug_xref(hemonc, compare_response, combo_drug):
    """Ensure that xrefs in combo treatments aren't included.

    See https://github.com/cancervariants/therapy-normalization/issues/417
    """
    response = hemonc.search("aspirin and dipyridamole")
    compare_response(response, MatchType.LABEL, combo_drug)

    response = hemonc.search("rxcui:226716")
    assert response.match_type == MatchType.NO_MATCH


def test_metadata(hemonc):
    """Test that source metadata returns correctly."""
    response = hemonc.search("search")
    assert response.source_meta_.data_license == "CC BY 4.0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/licenses/by/4.0/legalcode"
    )
    assert isodate.parse_date(response.source_meta_.version)
    assert (
        response.source_meta_.data_url
        == "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/9CY9C6"
    )
    assert response.source_meta_.rdp_url is None
    assert response.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "attribution": True,
        "share_alike": False,
    }



"""Construct hemonc test data."""

import csv
from pathlib import Path

from therapy.database import create_db
from therapy.etl import HemOnc

TEST_IDS = ["65", "105", "151", "26", "66258", "48"]

ho = HemOnc(create_db())  # don't need to write any data
ho._extract_data(False)
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "hemonc"

with ho._data_files.concepts.open() as f:
    concepts = list(csv.DictReader(f))
with ho._data_files.rels.open() as f:
    rels = list(csv.DictReader(f))
with ho._data_files.synonyms.open() as f:
    syns = list(csv.DictReader(f))

concept_ids = set()
for row in concepts:
    if row["concept_code"] in TEST_IDS:
        concept_ids.add(row["concept_code"])

test_rels_rows = []
rel_ids = set()
for row in rels:
    if row["concept_code_1"] in concept_ids:
        test_rels_rows.append(row)
        rel_ids.add(row["concept_code_2"])

rel_ids |= concept_ids
test_concepts_rows = []
for row in concepts:
    if row["concept_code"] in rel_ids:
        test_concepts_rows.append(row)

test_syn_rows = []
for row in syns:
    if row["synonym_concept_code"] in concept_ids:
        test_syn_rows.append(row)

with (TEST_DATA_DIR / ho._data_files.concepts.name).open("w") as f:
    writer = csv.DictWriter(f, concepts[0].keys())
    writer.writeheader()
    writer.writerows(test_concepts_rows)
with (TEST_DATA_DIR / ho._data_files.rels.name).open("w") as f:
    writer = csv.DictWriter(f, rels[0].keys())
    writer.writeheader()
    writer.writerows(test_rels_rows)
with (TEST_DATA_DIR / ho._data_files.synonyms.name).open("w") as f:
    writer = csv.DictWriter(f, syns[0].keys())
    writer.writeheader()
    writer.writerows(test_syn_rows)



"""Test the therapy querying method."""

import json
from datetime import datetime
from pathlib import Path

import pytest
from ga4gh.core.models import MappableConcept

from therapy.database.database import AbstractDatabase
from therapy.query import InvalidParameterError, QueryHandler
from therapy.schemas import MatchType, SourceName, Therapy


@pytest.fixture(scope="module")
def handler(database: AbstractDatabase):
    """Build query handler test fixture."""
    return QueryHandler(database)


@pytest.fixture(scope="module")
def search_handler(database: AbstractDatabase):
    """Build query handler test fixture."""

    class QueryGetter:
        def __init__(self):
            self.query_handler = QueryHandler(database)

        def search(self, query_str, incl="", excl="", infer=True):
            return self.query_handler.search(
                query_str=query_str, incl=incl, excl=excl, infer=infer
            )

    return QueryGetter()


@pytest.fixture(scope="module")
def normalize_handler(database: AbstractDatabase):
    """Provide Merge instance to test cases."""

    class QueryGetter:
        def __init__(self):
            self.query_handler = QueryHandler(database)

        def normalize(self, query_str, infer=True):
            return self.query_handler.normalize(query_str, infer)

        def normalize_unmerged(self, query_str, infer=True):
            return self.query_handler.normalize_unmerged(query_str, infer)

    return QueryGetter()


@pytest.fixture(scope="module")
def fixture_data(test_data: Path):
    """Fetch fixture data"""
    with (test_data / "fixtures" / "query_fixtures.json").open() as f:
        return json.load(f)


@pytest.fixture(scope="module")
def normalized_phenobarbital(fixture_data):
    """Create phenobarbital fixture."""
    return fixture_data["normalize_phenobarbital"]


@pytest.fixture(scope="module")
def normalized_cisplatin(fixture_data):
    """Create cisplatin fixture."""
    return fixture_data["normalize_cisplatin"]


@pytest.fixture(scope="module")
def normalized_spiramycin(fixture_data):
    """Create fixture for normalized spiramycin record."""
    return fixture_data["normalize_spiramycin"]


@pytest.fixture(scope="module")
def normalized_therapeutic_procedure(fixture_data):
    """Create a fixture for the Therapeutic Procedure concept. Used to validate
    single-member concept groups for the normalize endpoint.
    """
    return fixture_data["normalize_therapeutic_procedure"]


@pytest.fixture(scope="module")
def unmerged_normalized_cisplatin(fixture_data):
    """Create fixture data for unmerged normalized cisplatin record."""
    return fixture_data["unmerged_normalize_cisplatin"]


@pytest.fixture(scope="module")
def unmerged_normalized_therapeutic_procedure(fixture_data):
    """Create fixture data for unmerged normalized Therapeutic Procedure record.
    Tests whether responses are correctly formed from non-merged normalized result.
    """
    return fixture_data["unmerged_normalize_therapeutic_procedure"]


@pytest.fixture(scope="module")
def unmerged_normalized_spiramycin(fixture_data):
    """Create fixture data for spiramycin. Tests that responses are correctly formed
    from normalized results that only include some (not all) sources.
    """
    return fixture_data["unmerged_normalize_spiramycin"]


def compare_ta(response, fixture, query, match_type, warnings=None):
    """Verify correctness of returned core therapy object against test fixture

    :param Dict response: actual response
    :param Dict fixture: expected therapy object
    :param str query: query used in search
    :param MatchType match_type: expected MatchType
    :param List warnings: expected warnings
    """
    if warnings is None:
        warnings = []

    assert response.query == query

    # check warnings
    if warnings:
        assert len(response.warnings) == len(warnings), "warnings len"
        for e_warnings in warnings:
            for r_warnings in response.warnings:
                for e_key, e_val in e_warnings.items():
                    for r_val in r_warnings.values():
                        if e_key == r_val:
                            if isinstance(e_val, list):
                                assert set(r_val) == set(e_val), "warnings val"
                            else:
                                assert r_val == e_val, "warnings val"
    else:
        assert response.warnings == [], "warnings != []"

    assert response.match_type == match_type

    fixture = MappableConcept(**fixture.copy())
    assert (
        response.therapy.primaryCode.root == fixture.id.split("normalize.therapy.")[-1]
    )
    actual = response.therapy
    actual_keys = actual.model_dump(exclude_none=True).keys()
    fixture_keys = fixture.model_dump(exclude_none=True).keys()
    assert actual_keys == fixture_keys

    assert actual.id == fixture.id
    assert actual.conceptType == fixture.conceptType
    assert actual.label == fixture.label

    assert bool(actual.mappings) == bool(fixture.mappings)
    if actual.mappings:
        no_matches = []
        for actual_mapping in actual.mappings:
            match = None
            for fixture_mapping in fixture.mappings:
                if actual_mapping == fixture_mapping:
                    match = actual_mapping
                    break
            if not match:
                no_matches.append(actual_mapping)
        assert no_matches == [], no_matches
        assert len(actual.mappings) == len(fixture.mappings)

    def get_extension(extensions, name):
        matches = [e for e in extensions if e.name == name]
        if len(matches) > 1:
            pytest.fail(f"Multiple extensions named {name}")
        elif len(matches) == 1:
            return matches[0]
        else:
            return None

    assert bool(actual.extensions) == bool(fixture.extensions)
    if actual.extensions:
        ext_actual = actual.extensions
        ext_fixture = fixture.extensions

        approv_actual = get_extension(ext_actual, "regulatory_approval")
        approv_fixture = get_extension(ext_fixture, "regulatory_approval")
        assert (approv_actual is None) == (
            approv_fixture is None
        ), "regulatory_approval"
        if approv_actual and approv_fixture:
            ratings_actual = approv_actual.value.get("approval_ratings")
            ratings_fixture = approv_fixture.value.get("approval_ratings")
            if ratings_actual or ratings_fixture:
                assert set(ratings_actual) == set(ratings_fixture)
            assert set(approv_actual.value.get("approval_year", [])) == set(
                approv_fixture.value.get("approval_year", [])
            )
            approv_inds = [
                json.dumps(ind, sort_keys=True)
                for ind in approv_actual.value.get("has_indication", [])
            ]
            fixture_inds = [
                json.dumps(ind, sort_keys=True)
                for ind in approv_fixture.value.get("has_indication", [])
            ]
            assert set(approv_inds) == set(fixture_inds)

        tn_actual = get_extension(ext_actual, "trade_names")
        tn_fixture = get_extension(ext_fixture, "trade_names")
        assert (tn_actual is None) == (tn_fixture is None)
        if tn_fixture:
            assert tn_actual
            assert tn_actual.value
            assert set(tn_actual.value) == set(tn_fixture.value)

        fda_actual = get_extension(ext_actual, "fda_approval")
        fda_fixture = get_extension(ext_fixture, "fda_approval")
        assert (fda_actual is None) == (fda_fixture is None)
        if fda_fixture:
            assert fda_actual
            assert fda_actual.value.get("approval_status") == fda_fixture.value.get(
                "approval_status"
            )
            assert set(fda_actual.value.get("approval_year", [])) == set(
                fda_fixture.value.get("approval_year", [])
            )
            assert set(fda_actual.value.get("has_indication", [])) == set(
                fda_fixture.value.get("has_indication", [])
            )


def compare_unmerged_response(
    actual, query, warnings, match_type, fixture, compare_records
):
    """Compare response from normalize unmerged endpoint to fixture."""
    assert actual.query == query
    assert actual.warnings == warnings
    assert actual.match_type == match_type
    assert actual.normalized_concept_id == fixture["normalized_concept_id"]

    for source, match in actual.source_matches.items():
        assert match.source_meta_  # check that it's there
        for record in match.records:
            concept_id = record.concept_id
            fixture_drug = None
            # get corresponding fixture record
            for drug in fixture["source_matches"][source.value]["records"]:
                if drug["concept_id"] == concept_id:
                    fixture_drug = Therapy(**drug)
                    break
            assert fixture_drug, f"Unable to find fixture for {concept_id}"
            compare_records(record, fixture_drug)


def test_search(search_handler):
    """Test that query returns properly-structured response."""
    resp = search_handler.search("cisplatin")
    assert resp.query == "cisplatin"
    matches = resp.source_matches
    assert isinstance(matches, dict)
    assert len(matches) == 9
    wikidata = matches[SourceName.WIKIDATA]
    assert len(wikidata.records) == 1
    wikidata_record = wikidata.records[0]
    assert wikidata_record.label == "cisplatin"

    # test for not including redundant records w/ same match_type
    resp = search_handler.search("penicillamine")
    matches = resp.source_matches[SourceName.RXNORM]
    assert len(matches.records) == 1


def test_search_sources(search_handler):
    """Test inclusion and exclusion of sources in query."""
    # test blank params
    resp = search_handler.search("cisplatin")
    assert set(resp.source_matches.keys()) == {
        "Wikidata",
        "ChEMBL",
        "NCIt",
        "DrugBank",
        "ChemIDplus",
        "RxNorm",
        "HemOnc",
        "GuideToPHARMACOLOGY",
        "DrugsAtFDA",
    }

    # test partial inclusion
    resp = search_handler.search("cisplatin", incl="chembl,ncit")
    assert set(resp.source_matches.keys()) == {"ChEMBL", "NCIt"}

    # test full inclusion
    sources = "chembl,ncit,drugbank,wikidata,rxnorm,chemidplus,hemonc,guidetopharmacology,drugsatfda"
    resp = search_handler.search("cisplatin", incl=sources, excl="")
    assert set(resp.source_matches.keys()) == {
        "Wikidata",
        "ChEMBL",
        "NCIt",
        "DrugBank",
        "ChemIDplus",
        "RxNorm",
        "HemOnc",
        "GuideToPHARMACOLOGY",
        "DrugsAtFDA",
    }

    # test partial exclusion
    resp = search_handler.search("cisplatin", excl="chemidplus")
    assert set(resp.source_matches.keys()) == {
        "Wikidata",
        "ChEMBL",
        "NCIt",
        "DrugBank",
        "RxNorm",
        "HemOnc",
        "GuideToPHARMACOLOGY",
        "DrugsAtFDA",
    }

    # test full exclusion
    sources = "chembl,wikidata,drugbank,ncit,rxnorm,chemidplus,hemonc,guidetopharmacology,drugsatfda"
    resp = search_handler.search("cisplatin", excl=sources)
    assert set(resp.source_matches.keys()) == set()

    # test case insensitive
    resp = search_handler.search("cisplatin", excl="ChEmBl")
    assert set(resp.source_matches.keys()) == {
        "Wikidata",
        "NCIt",
        "DrugBank",
        "ChemIDplus",
        "RxNorm",
        "HemOnc",
        "GuideToPHARMACOLOGY",
        "DrugsAtFDA",
    }

    resp = search_handler.search("cisplatin", incl="wIkIdAtA,cHeMbL")
    assert set(resp.source_matches.keys()) == {"Wikidata", "ChEMBL"}

    # test error on invalid source names
    with pytest.raises(InvalidParameterError):
        resp = search_handler.search("cisplatin", incl="chambl")

    # test error for supplying both incl and excl args
    with pytest.raises(InvalidParameterError):
        resp = search_handler.search("cisplatin", incl="chembl", excl="wikidata")


def test_infer_option(search_handler, normalize_handler):
    """Test infer_namespace boolean option"""
    # drugbank
    query = "DB01174"
    expected_warnings = [
        {
            "inferred_namespace": "drugbank",
            "adjusted_query": "drugbank:" + query,
            "alternate_inferred_matches": [],
        }
    ]

    response = search_handler.search(query)
    assert (
        response.source_matches[SourceName.DRUGBANK].match_type == MatchType.CONCEPT_ID
    )
    assert response.warnings == expected_warnings

    response = normalize_handler.normalize(query)
    assert response.match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    # ncit
    query = "c739"
    expected_warnings = [
        {
            "inferred_namespace": "ncit",
            "adjusted_query": "ncit:" + query,
            "alternate_inferred_matches": [],
        }
    ]

    response = search_handler.search(query)
    assert response.source_matches["NCIt"].match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    response = normalize_handler.normalize(query)
    assert response.match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    # chemidplus
    query = "15663-27-1"
    expected_warnings = [
        {
            "inferred_namespace": "chemidplus",
            "adjusted_query": "chemidplus:" + query,
            "alternate_inferred_matches": [],
        }
    ]

    response = search_handler.search(query)
    assert response.source_matches["ChemIDplus"].match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    response = normalize_handler.normalize(query)
    assert response.match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    # chembl
    query = "chembl11359"
    expected_warnings = [
        {
            "inferred_namespace": "chembl",
            "adjusted_query": "chembl:" + query,
            "alternate_inferred_matches": [],
        }
    ]

    response = search_handler.search(query)
    assert response.source_matches["ChEMBL"].match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    response = normalize_handler.normalize(query)
    assert response.match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    # wikidata
    query = "q412415"
    expected_warnings = [
        {
            "inferred_namespace": "wikidata",
            "adjusted_query": "wikidata:" + query,
            "alternate_inferred_matches": [],
        }
    ]

    response = search_handler.search(query)
    assert response.source_matches["Wikidata"].match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    response = normalize_handler.normalize(query)
    assert response.match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    # drugs@fda
    query = "ANDA075036"
    expected_warnings = [
        {
            "inferred_namespace": "drugsatfda.anda",
            "adjusted_query": "drugsatfda.anda:075036",
            "alternate_inferred_matches": [],
        }
    ]

    response = search_handler.search(query)
    assert response.source_matches["DrugsAtFDA"].match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    response = normalize_handler.normalize(query)
    assert response.match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    query = "nda018057"
    expected_warnings = [
        {
            "inferred_namespace": "drugsatfda.nda",
            "adjusted_query": "drugsatfda.nda:018057",
            "alternate_inferred_matches": [],
        }
    ]

    response = search_handler.search(query)
    assert response.source_matches["DrugsAtFDA"].match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    response = normalize_handler.normalize(query)
    assert response.match_type == MatchType.CONCEPT_ID
    assert response.warnings == expected_warnings

    # test disabling namespace inference
    query = "DB01174"
    response = normalize_handler.normalize(query, infer=False)
    assert response.query == query
    assert response.warnings == []
    assert "record" not in response
    assert response.match_type == MatchType.NO_MATCH


def test_query_normalize(
    normalize_handler,
    normalized_phenobarbital,
    normalized_cisplatin,
    normalized_spiramycin,
    normalized_therapeutic_procedure,
):
    """Test that the normalized concept endpoint handles queries correctly."""
    # test merged id match
    query = "rxcui:2555"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_cisplatin, query, MatchType.CONCEPT_ID)

    # test concept id match
    query = "chemidplus:50-06-6"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_phenobarbital, query, MatchType.CONCEPT_ID)

    query = "hemonc:105"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_cisplatin, query, MatchType.CONCEPT_ID)

    # test label match
    query = "Phenobarbital"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_phenobarbital, query, MatchType.LABEL)

    # test trade name match
    query = "Platinol"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_cisplatin, query, MatchType.TRADE_NAME)

    # test alias match
    query = "cis-ddp"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_cisplatin, query, MatchType.ALIAS)

    query = "Rovamycine"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_spiramycin, query, MatchType.ALIAS)

    # test normalized group with single member
    query = "any therapy"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_therapeutic_procedure, query, MatchType.ALIAS)

    # test that term with multiple possible resolutions resolves at highest
    # match
    query = "Cisplatin"
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_cisplatin, query, MatchType.TRADE_NAME)

    # test whitespace stripping
    query = "   Cisplatin "
    response = normalize_handler.normalize(query)
    compare_ta(response, normalized_cisplatin, query, MatchType.TRADE_NAME)

    # test no match
    query = "zzzz fake therapy zzzz"
    response = normalize_handler.normalize(query)
    assert response.query == query
    assert response.warnings == []
    assert "record" not in response
    assert response.match_type == MatchType.NO_MATCH


def test_unmerged_normalize(
    normalize_handler,
    compare_records,
    unmerged_normalized_spiramycin,
    unmerged_normalized_cisplatin,
    unmerged_normalized_therapeutic_procedure,
):
    """Test correctness of unmerged normalize endpoint."""
    # concept ID match
    query = "rxcui:2555"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.CONCEPT_ID,
        unmerged_normalized_cisplatin,
        compare_records,
    )

    query = "chembl:CHEMBL11359"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.CONCEPT_ID,
        unmerged_normalized_cisplatin,
        compare_records,
    )
    query = "ncit:C49236"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.CONCEPT_ID,
        unmerged_normalized_therapeutic_procedure,
        compare_records,
    )

    query = "rxcui:9991"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.CONCEPT_ID,
        unmerged_normalized_spiramycin,
        compare_records,
    )

    query = "ncit:c839"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.CONCEPT_ID,
        unmerged_normalized_spiramycin,
        compare_records,
    )

    # namespace infer
    query = "CHEMBL11359"
    response = normalize_handler.normalize_unmerged(query)
    warning = {
        "inferred_namespace": "chembl",
        "adjusted_query": "chembl:CHEMBL11359",
        "alternate_inferred_matches": [],
    }
    compare_unmerged_response(
        response,
        query,
        [warning],
        MatchType.CONCEPT_ID,
        unmerged_normalized_cisplatin,
        compare_records,
    )

    # label match
    query = "cisplatin"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.LABEL,
        unmerged_normalized_cisplatin,
        compare_records,
    )

    query = "therapeutic procedure"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.LABEL,
        unmerged_normalized_therapeutic_procedure,
        compare_records,
    )

    query = "spiramycin"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.LABEL,
        unmerged_normalized_spiramycin,
        compare_records,
    )

    # alias/xref/trade name/associated with
    query = "Platinol-aq"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.TRADE_NAME,
        unmerged_normalized_cisplatin,
        compare_records,
    )

    query = "ndc:0143-9504"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.ASSOCIATED_WITH,
        unmerged_normalized_cisplatin,
        compare_records,
    )

    query = "CIS-DDP"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.ALIAS,
        unmerged_normalized_cisplatin,
        compare_records,
    )

    query = "TREAT"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.ALIAS,
        unmerged_normalized_therapeutic_procedure,
        compare_records,
    )

    query = "umls:C0087111"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.ASSOCIATED_WITH,
        unmerged_normalized_therapeutic_procedure,
        compare_records,
    )

    query = "rovamycin"
    response = normalize_handler.normalize_unmerged(query)
    compare_unmerged_response(
        response,
        query,
        [],
        MatchType.ALIAS,
        unmerged_normalized_spiramycin,
        compare_records,
    )


def test_merged_meta(normalize_handler):
    """Test population of source and resource metadata in merged querying."""
    query = "phenobarbital"
    response = normalize_handler.normalize(query)
    meta_items = response.source_meta_
    assert SourceName.RXNORM in meta_items
    assert SourceName.WIKIDATA in meta_items
    assert SourceName.NCIT in meta_items
    assert SourceName.CHEMIDPLUS in meta_items

    query = "RP 5337"
    response = normalize_handler.normalize(query)
    meta_items = response.source_meta_
    assert SourceName.NCIT in meta_items
    assert SourceName.CHEMIDPLUS in meta_items


def test_service_meta(search_handler, normalize_handler):
    """Test service meta info in response."""
    query = "pheno"

    response = search_handler.search(query)
    service_meta = response.service_meta_
    assert service_meta.name == "thera-py"
    assert service_meta.version != "unknown"
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/therapy-normalization"

    response = normalize_handler.normalize(query)
    service_meta = response.service_meta_
    assert service_meta.name == "thera-py"
    assert service_meta.version != "unknown"
    assert isinstance(service_meta.response_datetime, datetime)
    assert service_meta.url == "https://github.com/cancervariants/therapy-normalization"


def test_broken_db_handling(normalize_handler):
    """Test that query fails gracefully if mission-critical DB references are
    broken.

    The test database includes an identity record (fake:00001) with a
    purposely-broken merge_ref field. This lookup should not raise an
    exception.
    """
    query = "fake:00001"
    assert normalize_handler.normalize(query)



"""Test that the therapy normalizer works as intended for the ChEMBL source."""

import json
from pathlib import Path

import pytest

from therapy.etl import ChEMBL
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def chembl(test_source):
    """Provide test chembl query endpoint"""
    return test_source(ChEMBL)


@pytest.fixture(scope="module")
def fixture_data(test_data: Path):
    """Fetch fixture data"""
    with (test_data / "fixtures" / "chembl_fixtures.json").open() as f:
        return json.load(f)


@pytest.fixture(scope="module")
def cisplatin(fixture_data):
    """Create a cisplatin drug fixture."""
    return Therapy(**fixture_data["cisplatin"])


@pytest.fixture(scope="module")
def l745870(fixture_data):
    """Create a L-745870 drug fixture."""
    return Therapy(**fixture_data["l745870"])


# Test with aliases and trade names > 20
@pytest.fixture(scope="module")
def aspirin(fixture_data):
    """Create an aspirin drug fixture."""
    return Therapy(**fixture_data["aspirin"])


@pytest.fixture(scope="module")
def rosiglitazone(fixture_data):
    """Create a rosiglitazone fixture for checking import of withdrawn flag."""
    return Therapy(**fixture_data["rosiglitazone"])


def test_concept_id_cisplatin(cisplatin, chembl, compare_records):
    """Test that cisplatin drug normalizes to correct drug concept
    as a CONCEPT_ID match.
    """
    response = chembl.search("chembl:CHEMBL11359")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chembl.search("CHEMBL11359")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chembl.search("chembl:chembl11359")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chembl.search("cHEmbl:chembl11359")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chembl.search("cHEmbl:CHEMBL11359")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)


def test_cisplatin_label(cisplatin, chembl, compare_records):
    """Test that cisplatin drug normalizes to correct drug concept
    as a LABEL match.
    """
    response = chembl.search("CISPLATIN")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 2
    if len(response.records[0].aliases) > len(response.records[1].aliases):
        ind = 0
    else:
        ind = 1
    compare_records(response.records[ind], cisplatin)

    response = chembl.search("cisplatin")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 2
    if len(response.records[0].aliases) > len(response.records[1].aliases):
        ind = 0
    else:
        ind = 1
    compare_records(response.records[ind], cisplatin)


def test_cisplatin_alias(cisplatin, chembl, compare_records):
    """Test that alias term normalizes to correct drug concept as an
    ALIAS match.
    """
    response = chembl.search("cis-diamminedichloroplatinum(II)")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = chembl.search("INT230-6 COMPONENT CISPLATIn")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)


def test_no_match(chembl):
    """Test that a term normalizes to correct drug concept as a NO match."""
    response = chembl.search("cisplati")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0

    # Test white space in between label
    response = chembl.search("L - 745870")
    assert response.match_type == MatchType.NO_MATCH

    # Test empty query
    response = chembl.search("")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0


def test_l745870_concept_id(l745870, chembl, compare_records):
    """Test that L-745870 drug normalizes to correct drug concept
    as a CONCEPT_ID match.
    """
    response = chembl.search("chembl:CHEMBL267014")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], l745870)

    response = chembl.search("CHEMBL267014")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], l745870)

    response = chembl.search("chembl:chembl267014")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], l745870)

    response = chembl.search("cHEmbl:chembl267014")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], l745870)

    response = chembl.search("cHEmbl:CHEMBL267014")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], l745870)


def test_l745870_label(l745870, chembl, compare_records):
    """Test that L-745870 drug normalizes to correct drug concept
    as a LABEL match.
    """
    response = chembl.search("L-745870")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], l745870)

    response = chembl.search("l-745870")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], l745870)


def test_aspirin_concept_id(aspirin, chembl, compare_records):
    """Test that L-745870 drug normalizes to correct drug concept
    as a CONCEPT_ID match.
    """
    response = chembl.search("chembl:CHEMBL25")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], aspirin)

    response = chembl.search("CHEMBL25")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], aspirin)

    response = chembl.search("chembl:chembl25")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], aspirin)

    response = chembl.search("cHEmbl:chembl25")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], aspirin)

    response = chembl.search("cHEmbl:CHEMBL25")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], aspirin)


def test_aspirin_label(aspirin, chembl, compare_records):
    """Test that L-745870 drug normalizes to correct drug concept
    as a LABEL match.
    """
    response = chembl.search("ASPIRIN")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], aspirin)

    response = chembl.search("aspirin")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], aspirin)


def test_rosiglitazone(rosiglitazone, chembl, compare_records):
    """Test rosiglitazone -- checks for presence of chembl_withdrawn rating."""
    response = chembl.search("chembl:CHEMBL843")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], rosiglitazone)


def test_meta_info(chembl):
    """Test that the meta field is correct."""
    response = chembl.search("cisplatin")
    assert response.source_meta_.data_license == "CC BY-SA 3.0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/licenses/by-sa/3.0/"
    )
    assert response.source_meta_.version == "33"
    assert response.source_meta_.data_url.startswith(
        "ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb"
    )
    assert response.source_meta_.rdp_url == "http://reusabledata.org/chembl.html"
    assert response.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "share_alike": True,
        "attribution": True,
    }



"""Test that the therapy normalizer works as intended for the
Wikidata source.
"""

import isodate
import pytest

from therapy.etl.wikidata import Wikidata
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def wikidata(test_source):
    """Provide test Wikidata query endpoint"""
    return test_source(Wikidata)


@pytest.fixture(scope="module")
def cisplatin():
    """Create a cisplatin drug fixture."""
    params = {
        "label": "cisplatin",
        "concept_id": "wikidata:Q412415",
        "aliases": [
            "CDDP",
            "Cis-DDP",
            "CIS-DDP",
            "cis-diamminedichloroplatinum(II)",
            "Platinol",
            "Platinol-AQ",
        ],
        "xrefs": [
            "chembl:CHEMBL11359",
            "drugbank:DB00515",
            "rxcui:2555",
            "chemidplus:15663-27-1",
            "iuphar.ligand:5343",
        ],
        "associated_with": ["pubchem.compound:5702198"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def interferon_alfacon_1():
    """Create an Interferon alfacon-1 drug fixture."""
    params = {
        "label": "interferon alfacon-1",
        "concept_id": "wikidata:Q15353101",
        "aliases": [
            "CIFN",
            "consensus interferon",
            "IFN Alfacon-1",
            "Interferon Consensus, Methionyl",
            "methionyl interferon consensus",
            "methionyl-interferon-consensus",
            "rCon-IFN",
            "Recombinant Consensus Interferon",
            "Recombinant methionyl human consensus interferon",
        ],
        "xrefs": [
            "chembl:CHEMBL1201557",
            "drugbank:DB00069",
            "rxcui:59744",
            "chemidplus:118390-30-0",
        ],
    }
    return Therapy(**params)


# Test drug that has > 20 aliases
@pytest.fixture(scope="module")
def d_methamphetamine():
    """Create a D-methamphetamine drug fixture."""
    params = {
        "label": "D-methamphetamine",
        "concept_id": "wikidata:Q191924",
        "aliases": [],
        "approval_ratings": [],
        "xrefs": [
            "chembl:CHEMBL1201201",
            "drugbank:DB01577",
            "rxcui:6816",
            "chemidplus:537-46-2",
        ],
        "associated_with": [
            "pubchem.compound:10836",
        ],
        "trade_names": [],
    }
    return Therapy(**params)


# Test removing duplicate (case sensitive) aliases
@pytest.fixture(scope="module")
def atropine():
    """Create an atropine drug fixture."""
    params = {
        "label": "atropine",
        "concept_id": "wikidata:Q26272",
        "aliases": [
            "dl-Hyoscyamine",
            "dl-tropyltropate",
            "Mydriasine",
            "Tropine tropate",
            "(±)-atropine",
            "(±)-hyoscyamine",
            "(+-)-Atropine",
            "(+-)-Hyoscyamine",
            "(+,-)-Tropyl tropate",
            "(3-Endo)-8-methyl-8-azabicyclo[3.2.1]oct-3-yl tropate",
            "[(1S,5R)-8-Methyl-8-azabicyclo[3.2.1]oct-3-yl] "
            "3-hydroxy-2-phenyl-propanoate",
            "8-Methyl-8-azabicyclo[3.2.1]oct-3-yl " "3-hydroxy-2-phenylpropanoate",
            "8-Methyl-8-azabicyclo[3.2.1]oct-3-yl tropate",
        ],
        "approval_ratings": [],
        "xrefs": [
            "drugbank:DB00572",
            "chemidplus:51-55-8",
            "rxcui:1223",
            "chembl:CHEMBL517712",
            "iuphar.ligand:320",
        ],
        "associated_with": [
            "pubchem.compound:174174",
        ],
        "trade_names": [],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def basiliximab():
    """Create basiliximab test fixture to demonstrate rule-based filtering."""
    return Therapy(
        label="basiliximab",
        concept_id="wikidata:Q418702",
        aliases=["SDZ-CHI-621", "CHI-621", "CHI621", "chimeric mouse-human antiCD25"],
        xrefs=[
            "drugbank:DB00074",
            "rxcui:196102",
            "chembl:CHEMBL1201439",
            "chemidplus:179045-86-4",
            "iuphar.ligand:6879",
        ],
    )


def test_cisplatin(cisplatin, wikidata, compare_records):
    """Test that cisplatin drug normalizes to correct drug concept."""
    response = wikidata.search("wikidata:Q412415")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("wiKIdata:Q412415")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("wiKIdata:q412415")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("cisplatin")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("Cisplatin")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("CDDP")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("cddp")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("chembl:CHEMBL11359")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("drugbank:DB00515")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("rxcui:2555")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("chemidplus:15663-27-1")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)

    response = wikidata.search("pubchem.compound:5702198")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], cisplatin)


def test_atropine(atropine, wikidata, compare_records):
    """Test that atropine drug normalizes to correct drug concept."""
    response = wikidata.search("wikidata:Q26272")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)

    response = wikidata.search("wiKIdata:Q26272")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)

    response = wikidata.search("wiKIdata:q26272")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)

    response = wikidata.search("atropine")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)

    response = wikidata.search("Atropine")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)

    response = wikidata.search("Mydriasine")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)

    response = wikidata.search("(±)-Hyoscyamine")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)

    response = wikidata.search("chemidplus:51-55-8")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)

    response = wikidata.search("pubchem.compound:174174")
    assert response.match_type == MatchType.ASSOCIATED_WITH
    assert len(response.records) == 1
    compare_records(response.records[0], atropine)


def test_interferon_alfacon_1(interferon_alfacon_1, wikidata, compare_records):
    """Test that interferon alfacon-1 drug normalizes to correct concept"""
    response = wikidata.search("wikidata:Q15353101")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], interferon_alfacon_1)

    response = wikidata.search("wiKIdata:Q15353101")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], interferon_alfacon_1)

    response = wikidata.search("wiKIdata:q15353101")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], interferon_alfacon_1)

    response = wikidata.search("Interferon alfacon-1")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], interferon_alfacon_1)

    response = wikidata.search("Interferon Alfacon-1")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], interferon_alfacon_1)


def test_d_methamphetamine(d_methamphetamine, wikidata, compare_records):
    """Test that d_methamphetamine drug normalizes to correct concept."""
    response = wikidata.search("wikidata:Q191924")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], d_methamphetamine)

    response = wikidata.search("wiKIdata:Q191924")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], d_methamphetamine)

    response = wikidata.search("wiKIdata:q191924")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], d_methamphetamine)

    response = wikidata.search("D-methamphetamine")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], d_methamphetamine)

    response = wikidata.search("d-methamphetamine")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], d_methamphetamine)

    response = wikidata.search("rxcui:6816")
    assert response.match_type == MatchType.XREF
    assert len(response.records) == 1
    compare_records(response.records[0], d_methamphetamine)


def test_basiliximab(basiliximab, wikidata, compare_records):
    """Test that basiliximab terms resolve correctly"""
    response = wikidata.search("wikidata:Q418702")
    assert response.match_type == MatchType.CONCEPT_ID
    assert len(response.records) == 1
    compare_records(response.records[0], basiliximab)

    response = wikidata.search("basiliximab")
    assert response.match_type == MatchType.LABEL
    assert len(response.records) == 1
    compare_records(response.records[0], basiliximab)

    response = wikidata.search("CHI621")
    assert response.match_type == MatchType.ALIAS
    assert len(response.records) == 1
    compare_records(response.records[0], basiliximab)

    response = wikidata.search("Ig gamma-1 chain C region")
    assert response.match_type == MatchType.NO_MATCH


def test_case_no_match(wikidata):
    """Test that a term normalizes to NO match."""
    response = wikidata.search("cisplati")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0

    # Test white space in between label
    response = wikidata.search("Interferon alfacon - 1")
    assert response.match_type == MatchType.NO_MATCH

    # Test empty query
    response = wikidata.search("")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0


def test_meta_info(wikidata):
    """Test that the meta field is correct."""
    response = wikidata.search("cisplatin")
    assert response.source_meta_.data_license == "CC0 1.0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/publicdomain/zero/1.0/"
    )
    assert isodate.parse_date(response.source_meta_.version)
    assert response.source_meta_.data_url is None
    assert not response.source_meta_.rdp_url
    assert response.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "share_alike": False,
        "attribution": False,
    }



"""ETL methods for the Drugs@FDA source."""

import json
import logging

from tqdm import tqdm

from therapy.etl.base import Base
from therapy.schemas import (
    ApprovalRating,
    NamespacePrefix,
    RecordParams,
    SourceMeta,
)

_logger = logging.getLogger(__name__)


class DrugsAtFDA(Base):
    """Class for Drugs@FDA ETL methods."""

    def _load_meta(self) -> None:
        """Add Drugs@FDA metadata."""
        meta = {
            "data_license": "CC0",
            "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/legalcode",
            "version": self._version,
            "data_url": "https://open.fda.gov/apis/drug/drugsfda/download/",
            "rdp_url": None,
            "data_license_attributes": {
                "non_commercial": False,
                "share_alike": False,
                "attribution": False,
            },
        }
        self.database.add_source_metadata(self._name, SourceMeta(**meta))

    def _get_marketing_status_rating(
        self, products: list, concept_id: str
    ) -> str | None:
        """Get approval status rating from products list.
        :param List products: list of individual FDA product objects
        :param str concept_id: FDA application concept ID, used in reporting error
            messages
        :return: approval_rating value if successful, None if ambiguous or unavailable
        """
        statuses_map = {
            "Discontinued": ApprovalRating.FDA_DISCONTINUED.value,
            "Prescription": ApprovalRating.FDA_PRESCRIPTION.value,
            "Over-the-counter": ApprovalRating.FDA_OTC.value,
            "None (Tentative Approval)": ApprovalRating.FDA_TENTATIVE.value,
        }
        statuses = [p["marketing_status"] for p in products]
        if not all(s == statuses[0] for s in statuses):
            msg = (
                f"Application {concept_id} has inconsistent marketing "
                f"statuses: {statuses}"
            )
            _logger.info(msg)
            return None
        return statuses_map.get(statuses[0])

    def _transform_data(self) -> None:
        """Prepare source data for loading into DB."""
        with self._data_file.open() as f:  # type: ignore
            data = json.load(f)["results"]

        for result in tqdm(data, ncols=80, disable=self._silent):
            if "products" not in result:
                continue
            products = result["products"]

            app_no = result["application_number"]
            if app_no.startswith("NDA"):
                namespace = NamespacePrefix.DRUGSATFDA_NDA.value
            elif app_no.startswith("ANDA"):
                namespace = NamespacePrefix.DRUGSATFDA_ANDA.value
            else:
                # ignore biologics license applications (tentative)
                continue
            concept_id = f"{namespace}:{app_no.split('NDA')[-1]}"
            therapy: RecordParams = {"concept_id": concept_id}

            rating = self._get_marketing_status_rating(products, concept_id)
            if rating:
                therapy["approval_ratings"] = [rating]

            brand_names = [p["brand_name"] for p in products]

            aliases = []
            if "openfda" in result:
                openfda = result["openfda"]
                brand_name = openfda.get("brand_name")
                if brand_name:
                    brand_names += brand_name

                substances = openfda.get("substance_name", [])
                n_substances = len(set(substances))
                if n_substances > 1:
                    # if ambiguous, store all as aliases
                    msg = (
                        f"Application {concept_id} has {n_substances} "
                        f"substance names: {substances}"
                    )
                    _logger.debug(msg)
                    aliases += substances
                elif n_substances == 1:
                    therapy["label"] = substances[0]

                generics = openfda.get("generic_name", [])
                n_generic = len(set(generics))
                if n_generic > 1:
                    aliases += generics
                elif n_generic == 1:
                    if n_substances == 0:
                        # there are about 300 cases of this
                        therapy["label"] = generics[0]
                    else:
                        aliases.append(generics[0])

                therapy["associated_with"] = []
                unii = openfda.get("unii")
                if unii:
                    unii_items = [f"{NamespacePrefix.UNII.value}:{u}" for u in unii]
                    therapy["associated_with"] += unii_items  # type: ignore
                spl = openfda.get("spl_id")
                if spl:
                    spl_items = [f"{NamespacePrefix.SPL.value}:{s}" for s in spl]
                    therapy["associated_with"] += spl_items  # type: ignore
                ndc = openfda.get("product_ndc")
                if ndc:
                    ndc_items = [f"{NamespacePrefix.NDC.value}:{n}" for n in ndc]
                    therapy["associated_with"] += ndc_items  # type: ignore

                rxcui = openfda.get("rxcui")
                if rxcui:
                    therapy["xrefs"] = [
                        f"{NamespacePrefix.RXNORM.value}:{r}" for r in rxcui
                    ]

            therapy["trade_names"] = brand_names
            therapy["aliases"] = aliases
            self._load_therapy(therapy)



"""Test NCIT source"""

import re

import pytest

from therapy.etl.ncit import NCIt
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def ncit(test_source):
    """Provide test ncit query endpoint"""
    return test_source(NCIt)


@pytest.fixture(scope="module")
def voglibose():
    """Create a voglibose drug fixture.."""
    params = {
        "label": "Voglibose",
        "concept_id": "ncit:C95221",
        "aliases": [
            "3,4-Dideoxy-4-((2-Hydroxy-1-(Hydroxymethyl)Ethyl)Amino)-2-C-(Hydroxymethyl)-D-Epi-Inositol",  # F401
            "A-71100",
            "AO-128",
            "Basen",
            "N-(1,3-Dihydroxy-2-Propyl)Valiolamine",
            "VOGLIBOSE",
        ],
        "xrefs": ["chemidplus:83480-29-9"],
        "associated_with": ["unii:S77P977AG8", "umls:C0532578"],
        "approval_ratings": None,
        "trade_names": [],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def apricoxib():
    """Create an apricoxib drug fixture."""
    params = {
        "label": "Apricoxib",
        "concept_id": "ncit:C74021",
        "aliases": [
            "APRICOXIB",
            "COX-2 Inhibitor TG01",
            "CS-706",
            "R-109339",
            "TG01",
            "TP2001",
        ],
        "xrefs": ["chemidplus:197904-84-0"],
        "associated_with": ["unii:5X5HB3VZ3Z", "umls:C1737955"],
        "approval_ratings": None,
        "trade_names": [],
    }
    return Therapy(**params)


# Test aliases > 20
@pytest.fixture(scope="module")
def trastuzumab():
    """Create a Trastuzumab drug fixture."""
    params = {
        "label": "Trastuzumab",
        "concept_id": "ncit:C1647",
        "aliases": [],
        "xrefs": ["chemidplus:180288-69-1"],
        "associated_with": ["umls:C0728747", "unii:P188ANX8CK"],
        "approval_ratings": None,
        "trade_names": [],
    }
    return Therapy(**params)


# Needed for MetaKB
@pytest.fixture(scope="module")
def therapeutic_procedure():
    """Create a fixture for the Therapeutic Procedure class."""
    params = {
        "concept_id": "ncit:C49236",
        "label": "Therapeutic Procedure",
        "aliases": [
            "any therapy",
            "any_therapy",
            "Therapeutic",
            "Therapeutic Interventions",
            "Therapeutic Method",
            "Therapeutic Technique",
            "therapy",
            "Therapy",
            "TREAT",
            "Treatment",
            "TX",
            "treatment",
            "treatment or therapy",
            "treatment_or_therapy",
        ],
        "trade_names": [],
        "xrefs": [],
        "associated_with": ["umls:C0087111"],
    }
    return Therapy(**params)


# test correct CHEBI naming
@pytest.fixture(scope="module")
def ivermectin():
    """Create a fixture for Ivermectin."""
    params = {
        "label_and_type": "ncit:c61796",
        "src_name": "NCIt",
        "item_type": "identity",
        "concept_id": "ncit:C61796",
        "label": "Ivermectin",
        "aliases": [
            "Stromectol",
            "IVERMECTIN",
            "Avermectin A1a, 5-O-demethyl-25-de(1-methylpropyl)-22,23-dihydro-25-(1-methylethyl)-",
            "Sklice",
        ],
        "xrefs": ["chemidplus:70288-86-7"],
        "associated_with": ["umls:C0022322", "unii:8883YP2R6D", "CHEBI:6078"],
    }
    return Therapy(**params)


def test_concept_id_match(
    ncit,
    compare_response,
    voglibose,
    apricoxib,
    trastuzumab,
    therapeutic_procedure,
    ivermectin,
):
    """Test that concept ID query resolves to correct record."""
    response = ncit.search("ncit:C95221")
    compare_response(response, MatchType.CONCEPT_ID, voglibose)

    response = ncit.search("NCIT:C95221")
    compare_response(response, MatchType.CONCEPT_ID, voglibose)

    response = ncit.search("NCIt:c95221")
    compare_response(response, MatchType.CONCEPT_ID, voglibose)

    response = ncit.search("C95221")
    compare_response(response, MatchType.CONCEPT_ID, voglibose)

    response = ncit.search("ncit:C74021")
    compare_response(response, MatchType.CONCEPT_ID, apricoxib)

    response = ncit.search("NCIt:C74021")
    compare_response(response, MatchType.CONCEPT_ID, apricoxib)

    response = ncit.search("ncit:c74021")
    compare_response(response, MatchType.CONCEPT_ID, apricoxib)

    response = ncit.search("C74021")
    compare_response(response, MatchType.CONCEPT_ID, apricoxib)

    response = ncit.search("ncit:C1647")
    compare_response(response, MatchType.CONCEPT_ID, trastuzumab)

    response = ncit.search("NCIT:C1647")
    compare_response(response, MatchType.CONCEPT_ID, trastuzumab)

    response = ncit.search("NCIt:c1647")
    compare_response(response, MatchType.CONCEPT_ID, trastuzumab)

    response = ncit.search("C1647")
    compare_response(response, MatchType.CONCEPT_ID, trastuzumab)

    response = ncit.search("ncit:C49236")
    compare_response(response, MatchType.CONCEPT_ID, therapeutic_procedure)

    response = ncit.search("NCIT:C61796")
    compare_response(response, MatchType.CONCEPT_ID, ivermectin)


def test_label_match(ncit, compare_response, voglibose, apricoxib, trastuzumab):
    """Test that label query resolves to correct record."""
    response = ncit.search("voglibose")
    compare_response(response, MatchType.LABEL, voglibose)

    response = ncit.search("voglibose")
    compare_response(response, MatchType.LABEL, voglibose)

    response = ncit.search("trastuzumab")
    compare_response(response, MatchType.LABEL, trastuzumab)

    response = ncit.search("Trastuzumab")
    compare_response(response, MatchType.LABEL, trastuzumab)

    response = ncit.search("Apricoxib")
    compare_response(response, MatchType.LABEL, apricoxib)

    response = ncit.search("APRICOXIB")
    compare_response(response, MatchType.LABEL, apricoxib)


def test_alias_match(ncit, compare_response, voglibose, apricoxib):
    """Test that alias query resolves to correct record."""
    response = ncit.search("BASEN")
    compare_response(response, MatchType.ALIAS, voglibose)

    response = ncit.search("AO-128")
    compare_response(response, MatchType.ALIAS, voglibose)

    response = ncit.search("COX-2 Inhibitor TG01")
    compare_response(response, MatchType.ALIAS, apricoxib)


def test_xref_match(ncit, compare_response, voglibose, apricoxib, trastuzumab):
    """Test that xref query resolves to correct record."""
    response = ncit.search("chemidplus:83480-29-9")
    compare_response(response, MatchType.XREF, voglibose)

    response = ncit.search("CHEMIDPLUS:197904-84-0")
    compare_response(response, MatchType.XREF, apricoxib)

    response = ncit.search("chemidplus:180288-69-1")
    compare_response(response, MatchType.XREF, trastuzumab)


def test_assoc_with_match(ncit, compare_response, voglibose, apricoxib, trastuzumab):
    """Test that associated_with query resolves to correct record."""
    response = ncit.search("unii:S77P977AG8")
    compare_response(response, MatchType.ASSOCIATED_WITH, voglibose)

    response = ncit.search("unii:5X5HB3VZ3Z")
    compare_response(response, MatchType.ASSOCIATED_WITH, apricoxib)

    response = ncit.search("unii:P188ANX8CK")
    compare_response(response, MatchType.ASSOCIATED_WITH, trastuzumab)


def test_no_match(ncit):
    """Test that a term normalizes to NO match."""
    response = ncit.search("voglibo")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0

    # Test white space in between label
    response = ncit.search("Volgibo")
    assert response.match_type == MatchType.NO_MATCH

    # Test empty query
    response = ncit.search("")
    assert response.match_type == MatchType.NO_MATCH
    assert len(response.records) == 0


def test_meta_info(ncit):
    """Test that the meta field is correct."""
    response = ncit.search("NCIt")
    assert response.source_meta_.data_license == "CC BY 4.0"
    assert (
        response.source_meta_.data_license_url
        == "https://creativecommons.org/licenses/by/4.0/legalcode"
    )
    assert re.match(r"[0-9][0-9]\.[0-9][0-9][a-z]", response.source_meta_.version)
    assert response.source_meta_.data_url.startswith(
        "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/"
    )
    assert response.source_meta_.rdp_url == "http://reusabledata.org/ncit.html"
    assert response.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "share_alike": False,
        "attribution": True,
    }



"""Construct static mapping of disease normalizer input.
Assumes complete and functioning disease normalizer endpoint is available.
"""

import json
from pathlib import Path

from disease.database import create_db as create_disease_db
from disease.query import QueryHandler as DiseaseQueryHandler
from disease.schemas import NormalizationService as DiseaseNormalizationService

from therapy.database.dynamodb import DynamoDatabase
from therapy.etl import ChEMBL, HemOnc

TEST_ROOT = Path(__file__).resolve().parents[1]
TEST_DATA_DIRECTORY = TEST_ROOT / "data"


class ReadOnlyDatabase(DynamoDatabase):
    """Provide read-only instance of database for security's sake"""

    def add_record(self, record: dict, record_type: str = "identity") -> None:
        """Add new record to database"""


db = ReadOnlyDatabase()
disease_normalizer_table = {}


class SaveQueryHandler(DiseaseQueryHandler):
    """Disease query class which saves lookups and results"""

    def normalize(self, query: str) -> DiseaseNormalizationService:
        """Normalize query term"""
        response = super().normalize(query)
        result = response.normalized_id
        disease_normalizer_table[query.lower()] = result
        return response


disease_query_handler = SaveQueryHandler(create_disease_db())

ch = ChEMBL(database=db, data_path=TEST_DATA_DIRECTORY)
ch.disease_normalizer = disease_query_handler
ch.perform_etl(use_existing=True)

h = HemOnc(database=db, data_path=TEST_DATA_DIRECTORY)
h.disease_normalizer = disease_query_handler
h.perform_etl(use_existing=True)


with (TEST_DATA_DIRECTORY / "disease_normalization.json").open("w") as f:
    # for consistency/easier diffing
    sorted_dict = {  # noqa: C416
        key: value for key, value in sorted(disease_normalizer_table.items())
    }
    json.dump(sorted_dict, f, indent=2)



"""Provide DynamoDB client."""

import atexit
import logging
import sys
from collections.abc import Generator
from os import environ
from pathlib import Path

import boto3
import click
from boto3.dynamodb.conditions import Key
from botocore.exceptions import ClientError

from therapy import ITEM_TYPES, PREFIX_LOOKUP
from therapy.database.database import (
    AWS_ENV_VAR_NAME,
    SKIP_AWS_DB_ENV_NAME,
    VALID_AWS_ENV_NAMES,
    AbstractDatabase,
    AwsEnvName,
    DatabaseInitializationError,
    DatabaseReadError,
    DatabaseWriteError,
    confirm_aws_db_use,
)
from therapy.schemas import (
    RXNORM_BRAND_ITEM_TYPE,
    RecordType,
    RefType,
    SourceMeta,
    SourceName,
)

_logger = logging.getLogger(__name__)


class DynamoDatabase(AbstractDatabase):
    """Database class employing DynamoDB."""

    def __init__(self, db_url: str | None = None, **db_args) -> None:
        """Initialize Database class.

        :param db_url: URL endpoint for DynamoDB source
        :Keyword Arguments:
            * region_name: AWS region (defaults to "us-east-2")
        :raise DatabaseInitializationError: if initial setup fails
        """
        self.therapy_table = environ.get("THERAPY_DYNAMO_TABLE", "therapy_normalizer")
        region_name = db_args.get("region_name", "us-east-2")

        if AWS_ENV_VAR_NAME in environ:
            if "THERAPY_TEST" in environ:
                msg = f"Cannot have both THERAPY_TEST and {AWS_ENV_VAR_NAME} set."
                raise DatabaseInitializationError(msg)

            aws_env = environ[AWS_ENV_VAR_NAME]
            if aws_env not in VALID_AWS_ENV_NAMES:
                msg = f"{AWS_ENV_VAR_NAME} must be one of {VALID_AWS_ENV_NAMES}"
                raise DatabaseInitializationError(msg)

            skip_confirmation = environ.get(SKIP_AWS_DB_ENV_NAME)
            if (not skip_confirmation) or (
                skip_confirmation and skip_confirmation != "true"
            ):
                confirm_aws_db_use(environ[AWS_ENV_VAR_NAME])

            boto_params = {"region_name": region_name}

            if aws_env == AwsEnvName.DEVELOPMENT:
                self.therapy_table = environ.get(
                    "THERAPY_DYNAMO_TABLE", "therapy_normalizer_nonprod"
                )
        else:
            if db_url:
                endpoint_url = db_url
            elif "THERAPY_NORM_DB_URL" in environ:
                endpoint_url = environ["THERAPY_NORM_DB_URL"]
            else:
                endpoint_url = "http://localhost:8000"
            click.echo(f"***Using Therapy Database Endpoint: {endpoint_url}***")
            boto_params = {"region_name": region_name, "endpoint_url": endpoint_url}

        self.dynamodb = boto3.resource("dynamodb", **boto_params)
        self.dynamodb_client = boto3.client("dynamodb", **boto_params)

        # Only create tables for local instance
        envs_do_not_create_tables = {AWS_ENV_VAR_NAME, "THERAPY_TEST"}
        if not set(envs_do_not_create_tables) & set(environ):
            self.initialize_db()

        self.therapies = self.dynamodb.Table(self.therapy_table)
        self.batch = self.therapies.batch_writer()
        self._cached_sources: dict[str, SourceMeta] = {}
        atexit.register(self.close_connection)

    def list_tables(self) -> list[str]:
        """Return names of tables in database.

        :return: Table names in DynamoDB
        """
        return self.dynamodb_client.list_tables()["TableNames"]

    def drop_db(self) -> None:
        """Delete all tables from database. Requires manual confirmation.

        :raise DatabaseWriteError: if called in a protected setting with confirmation
            silenced.
        """
        try:
            if not self._check_delete_okay():
                return
        except DatabaseWriteError as e:
            raise e

        if self.therapy_table in self.list_tables():
            self.dynamodb.Table(self.therapy_table).delete()

    def _create_therapies_table(self) -> None:
        """Create Therapies table."""
        self.dynamodb.create_table(
            TableName=self.therapy_table,
            KeySchema=[
                {"AttributeName": "label_and_type", "KeyType": "HASH"},  # Partition key
                {"AttributeName": "concept_id", "KeyType": "RANGE"},  # Sort key
            ],
            AttributeDefinitions=[
                {"AttributeName": "label_and_type", "AttributeType": "S"},
                {"AttributeName": "concept_id", "AttributeType": "S"},
                {"AttributeName": "src_name", "AttributeType": "S"},
                {"AttributeName": "item_type", "AttributeType": "S"},
            ],
            GlobalSecondaryIndexes=[
                {
                    "IndexName": "src_index",
                    "KeySchema": [{"AttributeName": "src_name", "KeyType": "HASH"}],
                    "Projection": {"ProjectionType": "KEYS_ONLY"},
                    "ProvisionedThroughput": {
                        "ReadCapacityUnits": 10,
                        "WriteCapacityUnits": 10,
                    },
                },
                {
                    "IndexName": "item_type_index",
                    "KeySchema": [{"AttributeName": "item_type", "KeyType": "HASH"}],
                    "Projection": {"ProjectionType": "KEYS_ONLY"},
                    "ProvisionedThroughput": {
                        "ReadCapacityUnits": 10,
                        "WriteCapacityUnits": 10,
                    },
                },
            ],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

    def check_schema_initialized(self) -> bool:
        """Check if database schema is properly initialized.

        :return: True if DB appears to be fully initialized, False otherwise
        """
        existing_tables = self.list_tables()
        exists = self.therapy_table in existing_tables
        if not exists:
            _logger.info("%s table is missing or unavailable.", self.therapy_table)
        return exists

    def check_tables_populated(self) -> bool:
        """Perform rudimentary checks to see if tables are populated.

        Emphasis is on rudimentary -- if some rogueish element has deleted half of the
        therapy aliases, this method won't pick it up. It just wants to see if a few
        critical tables have at least a small number of records.

        :return: True if queries successful, false if DB appears empty
        """
        sources = self.therapies.query(
            IndexName="item_type_index",
            KeyConditionExpression=Key("item_type").eq("source"),
        ).get("Items", [])
        if len(sources) < len(SourceName):
            _logger.info("Therapy sources table is missing expected sources.")
            return False

        records = self.therapies.query(
            IndexName="item_type_index",
            KeyConditionExpression=Key("item_type").eq("identity"),
            Limit=1,
        )
        if len(records.get("Items", [])) < 1:
            _logger.info("Therapy records index is empty.")
            return False

        normalized_records = self.therapies.query(
            IndexName="item_type_index",
            KeyConditionExpression=Key("item_type").eq(RecordType.MERGER.value),
            Limit=1,
        )
        if len(normalized_records.get("Items", [])) < 1:
            _logger.info("Normalized therapy records index is empty.")
            return False

        return True

    def initialize_db(self) -> None:
        """Create therapy_normalizer table if not already created."""
        if not self.check_schema_initialized():
            self._create_therapies_table()

    def get_source_metadata(self, src_name: str | SourceName) -> SourceMeta | None:
        """Get license, versioning, data lookup, etc information for a source.

        :param src_name: name of the source to get data for
        :return: source metadata object if available
        """
        if isinstance(src_name, SourceName):
            src_name = src_name.value
        if src_name in self._cached_sources:
            return self._cached_sources[src_name]
        pk = f"{src_name.lower()}##source"
        concept_id = f"source:{src_name.lower()}"
        metadata = self.therapies.get_item(
            Key={"label_and_type": pk, "concept_id": concept_id}
        ).get("Item")
        if not metadata:
            return None
        formatted_metadata = SourceMeta(**metadata)
        self._cached_sources[src_name] = formatted_metadata
        return formatted_metadata

    def get_record_by_id(
        self, concept_id: str, case_sensitive: bool = True, merge: bool = False
    ) -> dict | None:
        """Fetch record corresponding to provided concept ID

        :param concept_id: concept ID for therapy record
        :param case_sensitive: if true, performs exact lookup, which is more efficient.
            Otherwise, performs filter operation, which doesn't require correct casing.
        :param merge: if true, look for merged record; look for identity record
            otherwise.
        :return: complete therapy record, if match is found; None otherwise
        """
        try:
            if merge:
                pk = f"{concept_id.lower()}##{RecordType.MERGER.value}"
            else:
                pk = f"{concept_id.lower()}##{RecordType.IDENTITY.value}"
            if case_sensitive:
                match = self.therapies.get_item(
                    Key={"label_and_type": pk, "concept_id": concept_id}
                )
                return match["Item"]
            exp = Key("label_and_type").eq(pk)
            response = self.therapies.query(KeyConditionExpression=exp)
            record = response["Items"][0]
            del record["label_and_type"]
            return record
        except ClientError as e:
            _logger.error(
                "boto3 client error on get_records_by_id for search term %s: %s",
                concept_id,
                e.response["Error"]["Message"],
            )
            return None
        except (KeyError, IndexError):  # record doesn't exist
            return None

    def get_refs_by_type(self, search_term: str, ref_type: RefType) -> list[str]:
        """Retrieve concept IDs for records matching the user's query. Other methods
        are responsible for actually retrieving full records.

        :param search_term: string to match against
        :param ref_type: type of match to look for.
        :return: list of associated concept IDs. Empty if lookup fails.
        """
        pk = f"{search_term}##{ref_type.value.lower()}"
        filter_exp = Key("label_and_type").eq(pk)
        try:
            matches = self.therapies.query(KeyConditionExpression=filter_exp)
            return [m["concept_id"] for m in matches.get("Items", None)]
        except ClientError as e:
            _logger.error(
                "boto3 client error on get_refs_by_type for search term %s: %s",
                search_term,
                e.response["Error"]["Message"],
            )
            return []

    def get_rxnorm_id_by_brand(self, brand_id: str) -> str | None:
        """Given RxNorm brand ID, retrieve associated drug concept ID.

        :param brand_id: rxcui brand identifier to dereference
        :return: RxNorm therapy concept ID if successful, None otherwise
        """
        pk = f"{brand_id.lower()}##{RXNORM_BRAND_ITEM_TYPE}"
        filter_exp = Key("label_and_type").eq(pk)
        try:
            matches = self.therapies.query(KeyConditionExpression=filter_exp)
        except ClientError as e:
            _logger.error(
                "boto3 client error on rx_brand fetch for brand ID {brand_id}: {e.response['Error']['Message']}",
                brand_id,
                e.response["Error"]["Message"],
            )
            return None
        if matches.get("Items") and len(matches["Items"]) == 1:
            return matches["Items"][0]["concept_id"]
        return None

    def get_drugsatfda_from_unii(self, unii: str) -> set[str]:
        """Get Drugs@FDA IDs associated with a single UNII, given that UNII. Used
        in merged concept generation.

        :param unii: UNII to find associations for
        :return: set of directly associated Drugs@FDA concept IDs.
        """
        dafda_concepts = set()
        associated_concepts = self.get_refs_by_type(unii, RefType.ASSOCIATED_WITH)
        for concept_id in associated_concepts:
            if concept_id.startswith("drugsatfda"):
                record = self.get_record_by_id(concept_id)
                if not record:
                    continue
                uniis = [
                    a for a in record.get("associated_with", []) if a.startswith("unii")
                ]
                if len(uniis) == 1:
                    dafda_concepts.add(concept_id)
        return dafda_concepts

    def get_all_concept_ids(self) -> set[str]:
        """Retrieve concept IDs for use in generating normalized records.

        :return: List of concept IDs as strings.
        """
        last_evaluated_key = None
        concept_ids = []
        params = {
            "ProjectionExpression": "concept_id,item_type",
        }
        while True:
            if last_evaluated_key:
                response = self.therapies.scan(
                    ExclusiveStartKey=last_evaluated_key, **params
                )
            else:
                response = self.therapies.scan(**params)
            records = response["Items"]
            for record in records:
                if record["item_type"] == "identity":
                    concept_id = record["concept_id"]
                    if not concept_id.startswith("source"):
                        concept_ids.append(record["concept_id"])
            last_evaluated_key = response.get("LastEvaluatedKey")
            if not last_evaluated_key:
                break
        return set(concept_ids)

    def get_all_records(self, record_type: RecordType) -> Generator[dict, None, None]:
        """Retrieve all source or normalized records. Either return all source records,
        or all records that qualify as "normalized" (i.e., merged groups + source
        records that are otherwise ungrouped).

        For example,

        >>> from therapy.database import create_db
        >>> from therapy.schemas import RecordType
        >>> db = create_db()
        >>> for record in db.get_all_records(RecordType.MERGER):
        >>>     pass  # do something

        :param record_type: type of result to return
        :return: Generator that lazily provides records as they are retrieved
        """
        last_evaluated_key = None
        while True:
            if last_evaluated_key:
                response = self.therapies.scan(
                    ExclusiveStartKey=last_evaluated_key,
                )
            else:
                response = self.therapies.scan()
            records = response.get("Items", [])
            for record in records:
                incoming_record_type = record.get("item_type")
                if record_type == RecordType.IDENTITY:
                    if incoming_record_type == record_type:
                        yield record
                else:
                    if (
                        incoming_record_type == RecordType.IDENTITY
                        and not record.get("merge_ref")
                    ) or incoming_record_type == RecordType.MERGER:
                        yield record
            last_evaluated_key = response.get("LastEvaluatedKey")
            if not last_evaluated_key:
                break

    def add_source_metadata(self, src_name: SourceName, metadata: SourceMeta) -> None:
        """Add new source metadata entry.

        :param src_name: name of source
        :param data: known source attributes
        :raise DatabaseWriteError: if write fails
        """
        src_name_value = src_name.value
        metadata_item = metadata.model_dump()
        metadata_item["src_name"] = src_name_value
        metadata_item["label_and_type"] = f"{str(src_name_value).lower()}##source"
        metadata_item["concept_id"] = f"source:{str(src_name_value).lower()}"
        metadata_item["item_type"] = "source"
        try:
            self.therapies.put_item(Item=metadata_item)
        except ClientError as e:
            raise DatabaseWriteError(e) from e

    def add_rxnorm_brand(self, brand_id: str, record_id: str) -> None:
        """Add RxNorm brand association to an existing RxNorm concept.

        :param brand_id: ID of RxNorm brand concept
        :param record_id: ID of RxNorm drug concept
        """
        item = {
            "label_and_type": f"{brand_id.lower()}##{RXNORM_BRAND_ITEM_TYPE}",
            "concept_id": record_id,
            "src_name": SourceName.RXNORM.value,
            "item_type": RXNORM_BRAND_ITEM_TYPE,
        }
        try:
            self.batch.put_item(Item=item)
        except ClientError as e:
            _logger.error(
                "boto3 client error on add_rxnorm_brand for %s -> %s: %s",
                brand_id,
                record_id,
                e.response["Error"]["Message"],
            )

    def add_record(self, record: dict, src_name: SourceName) -> None:
        """Add new record to database.

        :param record: record to upload
        :param src_name: name of source for record
        """
        concept_id = record["concept_id"]
        record["src_name"] = src_name.value
        label_and_type = f"{concept_id.lower()}##identity"
        record["label_and_type"] = label_and_type
        record["item_type"] = "identity"
        try:
            self.batch.put_item(Item=record)
        except ClientError as e:
            _logger.error(
                "boto3 client error on add_record for %s: %s",
                concept_id,
                e.response["Error"]["Message"],
            )
        for attr_type, item_type in ITEM_TYPES.items():
            if attr_type in record:
                value = record.get(attr_type)
                if not value:
                    continue
                if isinstance(value, str):
                    items = [value.lower()]
                else:
                    items = list({item.lower() for item in value})
                for item in items:
                    self._add_ref_record(
                        item, record["concept_id"], item_type, src_name
                    )

    def add_merged_record(self, record: dict) -> None:
        """Add merged record to database.

        :param record: merged record to add
        """
        concept_id = record["concept_id"]
        id_prefix = concept_id.split(":")[0].lower()
        record["src_name"] = PREFIX_LOOKUP[id_prefix]
        label_and_type = f"{concept_id.lower()}##{RecordType.MERGER.value}"
        record["label_and_type"] = label_and_type
        record["item_type"] = RecordType.MERGER.value
        try:
            self.batch.put_item(Item=record)
        except ClientError as e:
            _logger.error(
                "boto3 client error on add_record for %s: %s",
                concept_id,
                e.response["Error"]["Message"],
            )

    def _add_ref_record(
        self, term: str, concept_id: str, ref_type: str, src_name: SourceName
    ) -> None:
        """Add auxiliary/reference record to database.

        :param term: referent term
        :param concept_id: concept ID to refer to
        :param ref_type: one of {'alias', 'label', 'xref',
            'associated_with'}
        :param src_name: name of source for record
        """
        label_and_type = f"{term.lower()}##{ref_type}"
        record = {
            "label_and_type": label_and_type,
            "concept_id": concept_id.lower(),
            "src_name": src_name.value,
            "item_type": ref_type,
        }
        try:
            self.batch.put_item(Item=record)
        except ClientError as e:
            _logger.error(
                "boto3 client error adding reference %s for %s with match type %s: %s",
                term,
                concept_id,
                ref_type,
                e.response["Error"]["Message"],
            )

    def update_merge_ref(self, concept_id: str, merge_ref: str) -> None:
        """Update the merged record reference of an individual record to a new value.

        :param concept_id: record to update
        :param merge_ref: new ref value
        :raise DatabaseWriteError: if attempting to update non-existent record
        """
        label_and_type = f"{concept_id.lower()}##identity"
        key = {"label_and_type": label_and_type, "concept_id": concept_id}
        update_expression = "set merge_ref=:r"
        update_values = {":r": merge_ref.lower()}
        condition_expression = "attribute_exists(label_and_type)"
        try:
            self.therapies.update_item(
                Key=key,
                UpdateExpression=update_expression,
                ExpressionAttributeValues=update_values,
                ConditionExpression=condition_expression,
            )
        except ClientError as e:
            code = e.response.get("Error", {}).get("Code")
            if code == "ConditionalCheckFailedException":
                msg = f"No such record exists for keys {label_and_type}, {concept_id}"
                raise DatabaseWriteError(msg) from e
            _logger.error(
                "boto3 client error in `database.update_record()`: %s",
                e.response["Error"]["Message"],
            )

    def delete_normalized_concepts(self) -> None:
        """Remove merged records from the database. Use when performing a new update
        of normalized data.

        :raise DatabaseReadError: if DB client requires separate read calls and
            encounters a failure in the process
        :raise DatabaseWriteError: if deletion call fails
        """
        while True:
            with self.therapies.batch_writer(
                overwrite_by_pkeys=["label_and_type", "concept_id"]
            ) as batch:
                try:
                    response = self.therapies.query(
                        IndexName="item_type_index",
                        KeyConditionExpression=Key("item_type").eq(
                            RecordType.MERGER.value
                        ),
                    )
                except ClientError as e:
                    raise DatabaseReadError(e) from e
                records = response["Items"]
                if not records:
                    break
                for record in records:
                    batch.delete_item(
                        Key={
                            "label_and_type": record["label_and_type"],
                            "concept_id": record["concept_id"],
                        }
                    )

    def delete_source(self, src_name: SourceName) -> None:
        """Delete all data for a source. Use when updating source data.

        :param src_name: name of source to delete
        :raise DatabaseReadError: if DB client requires separate read calls and
            encounters a failure in the process
        :raise DatabaseWriteError: if deletion call fails
        """
        while True:
            try:
                response = self.therapies.query(
                    IndexName="src_index",
                    KeyConditionExpression=Key("src_name").eq(src_name.value),
                )
            except ClientError as e:
                raise DatabaseReadError(e) from e
            records = response["Items"]
            if not records:
                break
            with self.therapies.batch_writer(
                overwrite_by_pkeys=["label_and_type", "concept_id"]
            ) as batch:
                for record in records:
                    try:
                        batch.delete_item(
                            Key={
                                "label_and_type": record["label_and_type"],
                                "concept_id": record["concept_id"],
                            }
                        )
                    except ClientError as e:
                        raise DatabaseWriteError(e) from e

    def complete_write_transaction(self) -> None:
        """Conclude transaction or batch writing if relevant."""
        self.batch.__exit__(*sys.exc_info())
        self.batch = self.therapies.batch_writer()

    def close_connection(self) -> None:
        """Perform any manual connection closure procedures if necessary."""
        self.batch.__exit__(*sys.exc_info())

    def load_from_remote(self, url: str | None = None) -> None:
        """Load DB from remote dump. Not available for DynamoDB database backend.

        :param url: remote location to retrieve gzipped dump file from
        """
        raise NotImplementedError

    def export_db(self, export_location: Path) -> None:
        """Dump DB to specified location. Not available for DynamoDB database backend.

        :param export_location: path to save DB dump at
        """
        raise NotImplementedError



"""Module for Guide to PHARMACOLOGY ETL methods."""

import csv
import html
import logging
import re
from typing import Any

from tqdm import tqdm
from wags_tails.guide_to_pharmacology import GtoPLigandPaths

from therapy.etl.base import Base, SourceFormatError
from therapy.schemas import ApprovalRating, NamespacePrefix, SourceMeta, SourceName

_logger = logging.getLogger(__name__)

TAG_PATTERN = re.compile("</?[a-zA-Z]+>")
PMID_PATTERN = re.compile(r"\[PMID:[ ]?\d+\]")


class GuideToPHARMACOLOGY(Base):
    """Class for Guide to PHARMACOLOGY ETL methods."""

    def _extract_data(self, use_existing: bool) -> None:
        """Acquire source data.

        This method is responsible for initializing an instance of a data handler and
        setting ``self._data_files`` and ``self._version``.

        :param bool use_existing: if True, don't try to fetch latest source data
        """
        data_files, self._version = self._data_source.get_latest(
            from_local=use_existing
        )
        self._data_files: GtoPLigandPaths = data_files  # type: ignore

    def _transform_data(self) -> None:
        """Transform Guide To PHARMACOLOGY data."""
        data: dict[str, Any] = {}
        self._transform_ligands(data)
        self._transform_ligand_id_mappings(data)

        for param in tqdm(data.values(), ncols=80, disable=self._silent):
            self._load_therapy(param)

    @staticmethod
    def _process_name(name: str) -> str:
        """Remove tags from GtoP name object.
        :param name: raw drug referent
        :return: cleaned name (may be unchanged)
        """
        return re.sub(TAG_PATTERN, "", name)

    def _transform_ligands(self, data: dict) -> None:
        """Transform ligands data file and add this data to `data`.

        :param dict data: Transformed data
        """
        with self._data_files.ligands.open() as f:
            rows = csv.reader(f, delimiter="\t")

            # check that file structure is the same
            next(rows)
            if next(rows) != [
                "Ligand ID",
                "Name",
                "Species",
                "Type",
                "Approved",
                "Withdrawn",
                "Labelled",
                "Radioactive",
                "PubChem SID",
                "PubChem CID",
                "UniProt ID",
                "Ensembl ID",
                "ChEMBL ID",
                "Ligand Subunit IDs",
                "Ligand Subunit Name",
                "Ligand Subunit UniProt IDs",
                "Ligand Subunit Ensembl IDs",
                "IUPAC name",
                "INN",
                "Synonyms",
                "SMILES",
                "InChIKey",
                "InChI",
                "GtoImmuPdb",
                "GtoMPdb",
                "Antibacterial",
            ]:
                msg = "GtoP ligands file contains missing or unrecognized columns. See FAQ in README for suggested resolution."
                raise SourceFormatError(msg)

            for row in rows:
                params: dict[str, list[str] | str] = {
                    "concept_id": f"{NamespacePrefix.GUIDETOPHARMACOLOGY.value}:{row[0]}",
                    "label": self._process_name(row[1]),
                    "src_name": SourceName.GUIDETOPHARMACOLOGY.value,
                }

                approval_rating = self._set_approval_rating(row[4], row[5])
                if approval_rating:
                    params["approval_ratings"] = [approval_rating]

                associated_with = []
                aliases = []
                if row[8]:
                    associated_with.append(
                        f"{NamespacePrefix.PUBCHEMSUBSTANCE.value}:{row[8]}"
                    )
                if row[9]:
                    associated_with.append(
                        f"{NamespacePrefix.PUBCHEMCOMPOUND.value}:{row[9]}"
                    )
                if row[10]:
                    associated_with.append(f"{NamespacePrefix.UNIPROT.value}:{row[10]}")
                if row[17]:
                    aliases.append(self._process_name(row[17]))  # IUPAC
                if row[18]:
                    # International Non-proprietary Name assigned by the WHO
                    aliases.append(self._process_name(row[18]))
                if row[19]:
                    # synonyms
                    synonyms = row[19].split("|")
                    for s in synonyms:
                        if "&" in s and ";" in s:
                            name_code = s[s.index("&") : s.index(";") + 1]
                            if name_code.lower() in ["&reg;", "&trade;"]:
                                # Remove trademark symbols to allow for search
                                s = s.replace(name_code, "")
                            s = html.unescape(s)
                        aliases.append(self._process_name(s))
                if row[21]:
                    associated_with.append(
                        f"{NamespacePrefix.INCHIKEY.value}:{row[21]}"
                    )

                if associated_with:
                    params["associated_with"] = associated_with
                if aliases:
                    params["aliases"] = aliases

                data[params["concept_id"]] = params

    @staticmethod
    def _get_xrefs(ref: str, namespace: str) -> list[str]:
        """Construct xrefs from raw string.
        :param ref: raw ref value (may need to be separated)
        :param namspace: namespace prefix to use
        :return: List (usually with just one member) of xref IDs
        """
        xrefs = []
        for split_ref in ref.split("|"):
            xref = f"{namespace}:{split_ref}"
            xrefs.append(xref)
        return xrefs

    def _transform_ligand_id_mappings(self, data: dict) -> None:
        """Transform ligand_id_mappings and add this data to `data`
        All ligands found in this file should already be in data

        :param dict data: Transformed data
        """
        with self._data_files.ligand_id_mapping.open() as f:
            rows = csv.reader(f, delimiter="\t")
            next(rows)
            if next(rows) != [
                "Ligand id",
                "Name",
                "Species",
                "Type",
                "PubChem SID",
                "PubChem CID",
                "ChEMBl ID",
                "Chebi ID",
                "UniProt id",
                "Ensembl ID",
                "IUPAC name",
                "INN",
                "CAS",
                "DrugBank ID",
                "Drug Central ID",
            ]:
                msg = "GtoP ligand mapping file contains missing or unrecognized columns. See FAQ in README for suggested resolution."
                raise SourceFormatError(msg)

            for row in rows:
                concept_id = f"{NamespacePrefix.GUIDETOPHARMACOLOGY.value}:{row[0]}"

                if concept_id not in data:
                    _logger.debug("%s not in ligands", concept_id)
                    continue
                params = data[concept_id]
                xrefs = []
                associated_with = params.get("associated_with", [])
                if row[6]:
                    xrefs += self._get_xrefs(row[6], NamespacePrefix.CHEMBL.value)
                if row[7]:
                    # CHEBI IDs are already namespaced
                    for chebi_id in row[7].split("|"):
                        associated_with.append(chebi_id)
                if row[8]:
                    associated_with += self._get_xrefs(
                        row[8], NamespacePrefix.UNIPROT.value
                    )
                if row[12]:
                    xrefs += self._get_xrefs(row[12], NamespacePrefix.CASREGISTRY.value)
                if row[13]:
                    xrefs += self._get_xrefs(row[13], NamespacePrefix.DRUGBANK.value)
                if row[14]:
                    associated_with += self._get_xrefs(
                        row[14], NamespacePrefix.DRUGCENTRAL.value
                    )
                if xrefs:
                    params["xrefs"] = xrefs
                if associated_with:
                    params["associated_with"] = associated_with

    def _set_approval_rating(self, approved: str, withdrawn: str) -> str | None:
        """Set approval rating value.

        :param str approved: The drug is or has in the past been approved for human
            clinical use by a regulatory agency
        :param str withdrawn: The drug is no longer approved for its original clinical
            use in one or more countries
        :return: Approval rating
        """
        if approved and not withdrawn:
            approval_rating: str | None = ApprovalRating.GTOPDB_APPROVED.value
        elif withdrawn:
            approval_rating = ApprovalRating.GTOPDB_WITHDRAWN.value
        else:
            approval_rating = None
        return approval_rating

    def _load_meta(self) -> None:
        """Load Guide to PHARMACOLOGY metadata to database."""
        meta = SourceMeta(
            data_license="CC BY-SA 4.0",
            data_license_url="https://creativecommons.org/licenses/by-sa/4.0/",
            version=self._version,
            data_url="https://www.guidetopharmacology.org/download.jsp",
            rdp_url=None,
            data_license_attributes={
                "non_commercial": False,
                "share_alike": True,
                "attribution": True,
            },
        )
        self.database.add_source_metadata(SourceName.GUIDETOPHARMACOLOGY, meta)



"""Test Guide to PHARMACOLOGY source."""

import re

import pytest

from therapy.etl.guidetopharmacology import GuideToPHARMACOLOGY
from therapy.schemas import MatchType, Therapy


@pytest.fixture(scope="module")
def guidetopharmacology(test_source):
    """Provide test gtop query endpoint"""
    return test_source(GuideToPHARMACOLOGY)


@pytest.fixture(scope="module")
def cisplatin():
    """Create a cisplatin test fixture."""
    params = {
        "concept_id": "iuphar.ligand:5343",
        "label": "cisplatin",
        "approval_ratings": ["gtopdb_approved"],
        "xrefs": ["chembl:CHEMBL11359", "chemidplus:15663-27-1", "drugbank:DB00515"],
        "associated_with": [
            "pubchem.substance:178102005",
            "pubchem.compound:441203",
            "CHEBI:27899",
        ],
        "aliases": ["Platinol"],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def arginine_vasotocin():
    """Create an arginine vasotocin test fixture"""
    params = {
        "concept_id": "iuphar.ligand:2169",
        "label": "arginine vasotocin",
        "xrefs": ["chemidplus:113-80-4"],
        "associated_with": [
            "pubchem.substance:135652004",
            "pubchem.compound:68649",
            "inchikey:OXDZADMCOWPSOC-ICBIOJHSSA-N",
        ],
        "aliases": [
            "L-cysteinyl-L-tyrosyl-(3S)-DL-isoleucyl-L-glutaminyl-L-asparagyl-L-cysteinyl-DL-prolyl-L-arginyl-glycinamide (1->6)-disulfide",
            "argiprestocin",
            "[Arg8]vasotocin",
            "AVT",
        ],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def phenobarbital():
    """Create Phenobarbital test fixture"""
    params = {
        "concept_id": "iuphar.ligand:2804",
        "label": "phenobarbital",
        "approval_ratings": ["gtopdb_approved"],
        "xrefs": ["chembl:CHEMBL40", "chemidplus:50-06-6", "drugbank:DB01174"],
        "associated_with": [
            "pubchem.substance:135650817",
            "pubchem.compound:4763",
            "CHEBI:8069",
            "inchikey:DDBREPKUVSBGFI-UHFFFAOYSA-N",
            "drugcentral:2134",
        ],
        "aliases": [
            "5-ethyl-5-phenyl-1,3-diazinane-2,4,6-trione",
            "fenobarbital",
            "Luminal",
            "phenobarb",
            "phenobarbital sodium",
            "phenobarbitone",
            "phenylethylbarbiturate",
        ],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def cisapride():
    """Create cisapride test fixture"""
    params = {
        "concept_id": "iuphar.ligand:240",
        "label": "cisapride",
        "approval_ratings": ["gtopdb_withdrawn"],
        "xrefs": ["chembl:CHEMBL1729", "chemidplus:81098-60-4", "drugbank:DB00604"],
        "associated_with": [
            "pubchem.substance:135650104",
            "pubchem.compound:2769",
            "CHEBI:151790",
            "inchikey:DCSUBABJRXZOMT-UHFFFAOYSA-N",
            "drugcentral:660",
        ],
        "aliases": [
            "4-amino-5-chloro-N-[1-[3-(4-fluorophenoxy)propyl]-3-methoxypiperidin-4-yl]-2-methoxybenzamide",
            "Prepulsid",
            "Propulsid",
        ],
    }
    return Therapy(**params)


@pytest.fixture(scope="module")
def rolipram():
    """Create rolipram test fixture.
    Checks for correct handling of multiple references under same namespace.
    """
    return Therapy(
        concept_id="iuphar.ligand:5260",
        label="rolipram",
        xrefs=[
            "chembl:CHEMBL63",
            "chemidplus:61413-54-5",
            "drugbank:DB04149",
            "drugbank:DB03606",
        ],
        aliases=[
            "4-[3-(cyclopentyloxy)-4-methoxyphenyl]pyrrolidin-2-one",
            "(R,S)-rolipram",
            "(±)-rolipram",
        ],
        associated_with=[
            "CHEBI:104872",
            "pubchem.substance:178101944",
            "pubchem.compound:5092",
            "inchikey:HJORMJIFDVBMOB-UHFFFAOYSA-N",
        ],
    )


def test_concept_id_match(
    guidetopharmacology,
    compare_response,
    cisplatin,
    arginine_vasotocin,
    phenobarbital,
    cisapride,
    rolipram,
):
    """Test that concept ID queries work correctly."""
    resp = guidetopharmacology.search("iuphar.ligand:5343")
    compare_response(resp, MatchType.CONCEPT_ID, cisplatin)

    resp = guidetopharmacology.search("iuphar.ligand:2169")
    compare_response(resp, MatchType.CONCEPT_ID, arginine_vasotocin)

    resp = guidetopharmacology.search("iuphar.ligand:2804")
    compare_response(resp, MatchType.CONCEPT_ID, phenobarbital)

    resp = guidetopharmacology.search("iuphar.ligand:240")
    compare_response(resp, MatchType.CONCEPT_ID, cisapride)

    resp = guidetopharmacology.search("iuphar.ligand:5260")
    compare_response(resp, MatchType.CONCEPT_ID, rolipram)


def test_label_match(
    guidetopharmacology,
    compare_response,
    cisplatin,
    arginine_vasotocin,
    phenobarbital,
    cisapride,
    rolipram,
):
    """Test that label queries work correctly."""
    resp = guidetopharmacology.search("cisplatin")
    compare_response(resp, MatchType.LABEL, cisplatin)

    resp = guidetopharmacology.search("arginine vasotocin")
    compare_response(resp, MatchType.LABEL, arginine_vasotocin)

    resp = guidetopharmacology.search("Phenobarbital")
    compare_response(resp, MatchType.LABEL, phenobarbital)

    resp = guidetopharmacology.search("cisapride")
    compare_response(resp, MatchType.LABEL, cisapride)

    resp = guidetopharmacology.search("rolipram")
    compare_response(resp, MatchType.LABEL, rolipram)


def test_alias_match(
    guidetopharmacology,
    compare_response,
    cisplatin,
    arginine_vasotocin,
    phenobarbital,
    cisapride,
    rolipram,
):
    """Test that alias queries work correctly."""
    resp = guidetopharmacology.search("platinol")
    compare_response(resp, MatchType.ALIAS, cisplatin)

    resp = guidetopharmacology.search("AVT")
    compare_response(resp, MatchType.ALIAS, arginine_vasotocin)

    resp = guidetopharmacology.search("5-ethyl-5-phenyl-1,3-diazinane-2,4,6-trione")
    compare_response(resp, MatchType.ALIAS, phenobarbital)

    resp = guidetopharmacology.search("Prepulsid")
    compare_response(resp, MatchType.ALIAS, cisapride)

    resp = guidetopharmacology.search("(R,S)-rolipram")
    compare_response(resp, MatchType.ALIAS, rolipram)


def test_xref_match(
    guidetopharmacology,
    compare_response,
    cisplatin,
    arginine_vasotocin,
    phenobarbital,
    cisapride,
    rolipram,
):
    """Test that xref queries work correctly."""
    resp = guidetopharmacology.search("chemidplus:15663-27-1")
    compare_response(resp, MatchType.XREF, cisplatin)

    resp = guidetopharmacology.search("chemidplus:113-80-4")
    compare_response(resp, MatchType.XREF, arginine_vasotocin)

    resp = guidetopharmacology.search("chembl:CHEMBL40")
    compare_response(resp, MatchType.XREF, phenobarbital)

    resp = guidetopharmacology.search("drugbank:DB00604")
    compare_response(resp, MatchType.XREF, cisapride)

    resp = guidetopharmacology.search("drugbank:DB03606")
    compare_response(resp, MatchType.XREF, rolipram)


def test_associated_with_match(
    guidetopharmacology,
    compare_response,
    cisplatin,
    arginine_vasotocin,
    phenobarbital,
    cisapride,
    rolipram,
):
    """Test that associated_with queries work correctly."""
    resp = guidetopharmacology.search("pubchem.substance:178102005")
    compare_response(resp, MatchType.ASSOCIATED_WITH, cisplatin)

    resp = guidetopharmacology.search("pubchem.compound:441203")
    compare_response(resp, MatchType.ASSOCIATED_WITH, cisplatin)

    resp = guidetopharmacology.search("pubchem.substance:135652004")
    compare_response(resp, MatchType.ASSOCIATED_WITH, arginine_vasotocin)

    resp = guidetopharmacology.search("pubchem.compound:68649")
    compare_response(resp, MatchType.ASSOCIATED_WITH, arginine_vasotocin)

    resp = guidetopharmacology.search("pubchem.substance:135650817")
    compare_response(resp, MatchType.ASSOCIATED_WITH, phenobarbital)

    resp = guidetopharmacology.search("pubchem.compound:4763")
    compare_response(resp, MatchType.ASSOCIATED_WITH, phenobarbital)

    resp = guidetopharmacology.search("drugcentral:2134")
    compare_response(resp, MatchType.ASSOCIATED_WITH, phenobarbital)

    resp = guidetopharmacology.search("pubchem.substance:135650104")
    compare_response(resp, MatchType.ASSOCIATED_WITH, cisapride)

    resp = guidetopharmacology.search("pubchem.compound:2769")
    compare_response(resp, MatchType.ASSOCIATED_WITH, cisapride)

    resp = guidetopharmacology.search("inchikey:DCSUBABJRXZOMT-UHFFFAOYSA-N")
    compare_response(resp, MatchType.ASSOCIATED_WITH, cisapride)

    resp = guidetopharmacology.search("pubchem.substance:178101944")
    compare_response(resp, MatchType.ASSOCIATED_WITH, rolipram)


def test_no_match(guidetopharmacology):
    """Test that no match queries work correctly."""
    resp = guidetopharmacology.search("178102005")
    assert resp.match_type == MatchType.NO_MATCH
    assert len(resp.records) == 0

    resp = guidetopharmacology.search("guidetopharmacology:5343")
    assert resp.match_type == MatchType.NO_MATCH
    assert len(resp.records) == 0

    resp = guidetopharmacology.search("")
    assert resp.match_type == MatchType.NO_MATCH
    assert len(resp.records) == 0


def test_meta_info(guidetopharmacology):
    """Test that metadata is correct."""
    resp = guidetopharmacology.search("search")
    assert resp.source_meta_.data_license == "CC BY-SA 4.0"
    assert (
        resp.source_meta_.data_license_url
        == "https://creativecommons.org/licenses/by-sa/4.0/"
    )
    assert re.match(r"\d{4}.\d+", resp.source_meta_.version)
    assert (
        resp.source_meta_.data_url == "https://www.guidetopharmacology.org/download.jsp"
    )
    assert resp.source_meta_.rdp_url is None
    assert resp.source_meta_.data_license_attributes == {
        "non_commercial": False,
        "attribution": True,
        "share_alike": True,
    }



"""Build drugsatfda test data"""

import json
from pathlib import Path

from therapy.database import create_db
from therapy.etl import DrugsAtFDA

TEST_IDS = [
    "NDA020221",
    "NDA022334",
    "NDA050682",
    "ANDA074656",
    "ANDA075036",
    "ANDA074735",
    "ANDA206774",
    "ANDA207323",
    "NDA018057",
    "ANDA072267",
    "NDA017604",
    "NDA210595",
    "NDA202450",
    "NDA091141",
    "NDA022007",
    "NDA050682",
    "NDA017604",
    "ANDA214475",
]

daf = DrugsAtFDA(create_db())
daf._extract_data(False)
TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "drugsatfda"
outfile_path = TEST_DATA_DIR / daf._data_file.name

with daf._data_file.open() as f:
    data = json.load(f)

out_data = {"meta": data["meta"], "results": []}

for record in data["results"]:
    if record["application_number"] in TEST_IDS:
        out_data["results"].append(record)

with outfile_path.open("w") as f:
    json.dump(out_data, f)



"""Provide database clients."""

from .database import (
    AWS_ENV_VAR_NAME,
    AbstractDatabase,
    DatabaseError,
    DatabaseInitializationError,
    DatabaseReadError,
    DatabaseWriteError,
    create_db,
)

__all__ = [
    "AWS_ENV_VAR_NAME",
    "AbstractDatabase",
    "DatabaseError",
    "DatabaseInitializationError",
    "DatabaseReadError",
    "DatabaseWriteError",
    "create_db",
]
